# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest
from datetime import datetime

import mock

from libcloud.common.aws import AWSRequestSignerAlgorithmV4
from libcloud.common.aws import SignedAWSConnection
from libcloud.common.aws import UNSIGNED_PAYLOAD
from libcloud.test import LibcloudTestCase


class EC2MockDriver(object):
    region_name = 'my_region'


class AWSRequestSignerAlgorithmV4TestCase(LibcloudTestCase):

    def setUp(self):
        SignedAWSConnection.driver = EC2MockDriver()
        SignedAWSConnection.service_name = 'my_service'
        SignedAWSConnection.version = '2013-10-15'
        self.connection = SignedAWSConnection('my_key', 'my_secret')

        self.signer = AWSRequestSignerAlgorithmV4(access_key='my_key',
                                                  access_secret='my_secret',
                                                  version='2013-10-15',
                                                  connection=self.connection)

        SignedAWSConnection.action = '/my_action/'
        SignedAWSConnection.driver = EC2MockDriver()

        self.now = datetime(2015, 3, 4, hour=17, minute=34, second=52)

    def test_v4_signature(self):
        params = {
            'Action': 'DescribeInstances',
            'Version': '2013-10-15'
        }
        headers = {
            'Host': 'ec2.eu-west-1.amazonaws.com',
            'Accept-Encoding': 'gzip,deflate',
            'X-AMZ-Date': '20150304T173452Z',
            'User-Agent': 'libcloud/0.17.0 (Amazon EC2 (eu-central-1)) '
        }
        dt = self.now
        sig = self.signer._get_authorization_v4_header(params=params,
                                                       headers=headers,
                                                       dt=dt,
                                                       method='GET',
                                                       path='/my_action/')
        self.assertEqual(sig, 'AWS4-HMAC-SHA256 '
                              'Credential=my_key/20150304/my_region/my_service/aws4_request, '
                              'SignedHeaders=accept-encoding;host;user-agent;x-amz-date, '
                              'Signature=f9868f8414b3c3f856c7955019cc1691265541f5162b9b772d26044280d39bd3')

    def test_v4_signature_contains_user_id(self):
        sig = self.signer._get_authorization_v4_header(params={}, headers={},
                                                       dt=self.now)
        self.assertIn('Credential=my_key/', sig)

    def test_v4_signature_contains_credential_scope(self):
        with mock.patch('libcloud.common.aws.AWSRequestSignerAlgorithmV4._get_credential_scope') as mock_get_creds:
            mock_get_creds.return_value = 'my_credential_scope'
            sig = self.signer._get_authorization_v4_header(params={}, headers={}, dt=self.now)

        self.assertIn('Credential=my_key/my_credential_scope, ', sig)

    def test_v4_signature_contains_signed_headers(self):
        with mock.patch('libcloud.common.aws.AWSRequestSignerAlgorithmV4._get_signed_headers') as mock_get_headers:
            mock_get_headers.return_value = 'my_signed_headers'
            sig = self.signer._get_authorization_v4_header({}, {}, self.now,
                                                           method='GET',
                                                           path='/')
        self.assertIn('SignedHeaders=my_signed_headers, ', sig)

    def test_v4_signature_contains_signature(self):
        with mock.patch('libcloud.common.aws.AWSRequestSignerAlgorithmV4._get_signature') as mock_get_signature:
            mock_get_signature.return_value = 'my_signature'
            sig = self.signer._get_authorization_v4_header({}, {}, self.now)
        self.assertIn('Signature=my_signature', sig)

    def test_get_signature_(self):
        def _sign(key, msg, hex=False):
            if hex:
                return 'H|%s|%s' % (key, msg)
            else:
                return '%s|%s' % (key, msg)

        with mock.patch('libcloud.common.aws.AWSRequestSignerAlgorithmV4._get_key_to_sign_with') as mock_get_key:
            with mock.patch('libcloud.common.aws.AWSRequestSignerAlgorithmV4._get_string_to_sign') as mock_get_string:
                with mock.patch('libcloud.common.aws._sign', new=_sign):
                    mock_get_key.return_value = 'my_signing_key'
                    mock_get_string.return_value = 'my_string_to_sign'
                    sig = self.signer._get_signature({}, {}, self.now,
                                                     method='GET', path='/', data=None)

        self.assertEqual(sig, 'H|my_signing_key|my_string_to_sign')

    def test_get_string_to_sign(self):
        with mock.patch('hashlib.sha256') as mock_sha256:
            mock_sha256.return_value.hexdigest.return_value = 'chksum_of_canonical_request'
            to_sign = self.signer._get_string_to_sign({}, {}, self.now,
                                                      method='GET', path='/', data=None)

        self.assertEqual(to_sign,
                         'AWS4-HMAC-SHA256\n'
                         '20150304T173452Z\n'
                         '20150304/my_region/my_service/aws4_request\n'
                         'chksum_of_canonical_request')

    def test_get_key_to_sign_with(self):
        def _sign(key, msg, hex=False):
            return '%s|%s' % (key, msg)

        with mock.patch('libcloud.common.aws._sign', new=_sign):
            key = self.signer._get_key_to_sign_with(self.now)

        self.assertEqual(key, 'AWS4my_secret|20150304|my_region|my_service|aws4_request')

    def test_get_signed_headers_contains_all_headers_lowercased(self):
        headers = {'Content-Type': 'text/plain', 'Host': 'my_host', 'X-Special-Header': ''}
        signed_headers = self.signer._get_signed_headers(headers)

        self.assertIn('content-type', signed_headers)
        self.assertIn('host', signed_headers)
        self.assertIn('x-special-header', signed_headers)

    def test_get_signed_headers_concats_headers_sorted_lexically(self):
        headers = {'Host': 'my_host', 'X-Special-Header': '', '1St-Header': '2', 'Content-Type': 'text/plain'}
        signed_headers = self.signer._get_signed_headers(headers)

        self.assertEqual(signed_headers, '1st-header;content-type;host;x-special-header')

    def test_get_credential_scope(self):
        scope = self.signer._get_credential_scope(self.now)
        self.assertEqual(scope, '20150304/my_region/my_service/aws4_request')

    def test_get_canonical_headers_joins_all_headers(self):
        headers = {
            'accept-encoding': 'gzip,deflate',
            'host': 'my_host',
        }
        self.assertEqual(self.signer._get_canonical_headers(headers),
                         'accept-encoding:gzip,deflate\n'
                         'host:my_host\n')

    def test_get_canonical_headers_sorts_headers_lexically(self):
        headers = {
            'accept-encoding': 'gzip,deflate',
            'host': 'my_host',
            '1st-header': '2',
            'x-amz-date': '20150304T173452Z',
            'user-agent': 'my-ua'
        }
        self.assertEqual(self.signer._get_canonical_headers(headers),
                         '1st-header:2\n'
                         'accept-encoding:gzip,deflate\n'
                         'host:my_host\n'
                         'user-agent:my-ua\n'
                         'x-amz-date:20150304T173452Z\n')

    def test_get_canonical_headers_lowercases_headers_names(self):
        headers = {
            'Accept-Encoding': 'GZIP,DEFLATE',
            'User-Agent': 'My-UA'
        }
        self.assertEqual(self.signer._get_canonical_headers(headers),
                         'accept-encoding:GZIP,DEFLATE\n'
                         'user-agent:My-UA\n')

    def test_get_canonical_headers_trims_header_values(self):
        # TODO: according to AWS spec (and RFC 2616 Section 4.2.) excess whitespace
        # from inside non-quoted strings should be stripped. Now we only strip the
        # start and end of the string. See
        # http://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html
        headers = {
            'accept-encoding': '   gzip,deflate',
            'user-agent': 'libcloud/0.17.0 '
        }
        self.assertEqual(self.signer._get_canonical_headers(headers),
                         'accept-encoding:gzip,deflate\n'
                         'user-agent:libcloud/0.17.0\n')

    def test_get_request_params_joins_params_sorted_lexically(self):
        self.assertEqual(self.signer._get_request_params({
            'Action': 'DescribeInstances',
            'Filter.1.Name': 'state',
            'Version': '2013-10-15'
        }),
            'Action=DescribeInstances&Filter.1.Name=state&Version=2013-10-15')

    def test_get_canonical_headers_allow_numeric_header_value(self):
        headers = {
            'Accept-Encoding': 'gzip,deflate',
            'Content-Length': 314
        }
        self.assertEqual(self.signer._get_canonical_headers(headers),
                         'accept-encoding:gzip,deflate\n'
                         'content-length:314\n')

    def test_get_request_params_allows_integers_as_value(self):
        self.assertEqual(self.signer._get_request_params({'Action': 'DescribeInstances', 'Port': 22}),
                         'Action=DescribeInstances&Port=22')

    def test_get_request_params_urlquotes_params_keys(self):
        self.assertEqual(self.signer._get_request_params({'Action+Reaction': 'DescribeInstances'}),
                         'Action%2BReaction=DescribeInstances')

    def test_get_request_params_urlquotes_params_values(self):
        self.assertEqual(self.signer._get_request_params({
            'Action': 'DescribeInstances&Addresses',
            'Port-Range': '2000 3000'
        }),
            'Action=DescribeInstances%26Addresses&Port-Range=2000%203000')

    def test_get_request_params_urlquotes_params_values_allows_safe_chars_in_value(self):
        # http://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html
        self.assertEqual('Action=a~b.c_d-e',
                         self.signer._get_request_params({'Action': 'a~b.c_d-e'}))

    def test_get_payload_hash_returns_digest_of_empty_string_for_GET_requests(self):
        SignedAWSConnection.method = 'GET'
        self.assertEqual(self.signer._get_payload_hash(method='GET'),
                         'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855')

    def test_get_payload_hash_with_data_for_PUT_requests(self):
        SignedAWSConnection.method = 'PUT'
        self.assertEqual(self.signer._get_payload_hash(method='PUT', data='DUMMY'),
                         'ceec12762e66397b56dad64fd270bb3d694c78fb9cd665354383c0626dbab013')

    def test_get_payload_hash_with_empty_data_for_POST_requests(self):
        SignedAWSConnection.method = 'POST'
        self.assertEqual(self.signer._get_payload_hash(method='POST'),
                         UNSIGNED_PAYLOAD)

    def test_get_canonical_request(self):
        req = self.signer._get_canonical_request(
            {'Action': 'DescribeInstances', 'Version': '2013-10-15'},
            {'Accept-Encoding': 'gzip,deflate', 'User-Agent': 'My-UA'},
            method='GET',
            path='/my_action/',
            data=None
        )
        self.assertEqual(req, 'GET\n'
                              '/my_action/\n'
                              'Action=DescribeInstances&Version=2013-10-15\n'
                              'accept-encoding:gzip,deflate\n'
                              'user-agent:My-UA\n'
                              '\n'
                              'accept-encoding;user-agent\n'
                              'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855')

    def test_post_canonical_request(self):
        req = self.signer._get_canonical_request(
            {'Action': 'DescribeInstances', 'Version': '2013-10-15'},
            {'Accept-Encoding': 'gzip,deflate', 'User-Agent': 'My-UA'},
            method='POST',
            path='/my_action/',
            data='{}'
        )
        self.assertEqual(req, 'POST\n'
                              '/my_action/\n'
                              'Action=DescribeInstances&Version=2013-10-15\n'
                              'accept-encoding:gzip,deflate\n'
                              'user-agent:My-UA\n'
                              '\n'
                              'accept-encoding;user-agent\n'
                              '44136fa355b3678a1146ad16f7e8649e94fb4fc21fe77e8310c060f61caaff8a')

if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Kamatera node driver
"""
import json
import datetime
import time

from libcloud.utils.py3 import basestring
from libcloud.compute.base import NodeDriver, NodeLocation, NodeSize
from libcloud.compute.base import NodeImage, Node, NodeState
from libcloud.compute.types import Provider
from libcloud.common.base import ConnectionUserAndKey, JsonResponse


class KamateraResponse(JsonResponse):
    """
    Response class for KamateraDriver
    """

    def parse_error(self):
        data = self.parse_body()
        if 'message' in data:
            return data['message']
        else:
            return json.dumps(data)


class KamateraConnection(ConnectionUserAndKey):
    """
    Connection class for KamateraDriver
    """

    host = 'cloudcli.cloudwm.com'
    responseCls = KamateraResponse

    def add_default_headers(self, headers):
        """Adds headers that are needed for all requests"""
        headers['AuthClientId'] = self.user_id
        headers['AuthSecret'] = self.key
        headers['Accept'] = 'application/json'
        headers['Content-Type'] = 'application/json'
        return headers


class KamateraNodeDriver(NodeDriver):
    """
    Kamatera node driver

    :keyword    key: API Client ID, required for authentication
    :type       key: ``str``

    :keyword    secret: API Secret, required for authentcaiont
    :type       secret: ``str``
    """

    type = Provider.KAMATERA
    name = 'Kamatera'
    website = 'https://www.kamatera.com/'
    connectionCls = KamateraConnection
    features = {'create_node': ['password', 'generates_password', 'ssh_key']}

    EX_BILLINGCYCLE_HOURLY = 'hourly'
    EX_BILLINGCYCLE_MONTHLY = 'monthly'

    def list_locations(self):
        """
        List available locations for deployment

        :rtype: ``list`` of :class:`NodeLocation`
        """
        response = self.connection.request('service/server?datacenter=1')
        return [self.ex_get_location(
            datacenter['id'], datacenter['subCategory'], datacenter['name']
        ) for datacenter in response.object]

    def list_sizes(self, location):
        """
        List predefined sizes for the given location.

        :param location: Location of the deployment.
        :type location: :class:`.NodeLocation`

        @inherits: :class:`NodeDriver.list_sizes`
        """
        response = self.connection.request(
            'service/server?sizes=1&datacenter=%s' % location.id)
        return [self.ex_get_size(
            size['ramMB'],
            size['diskSizeGB'],
            size['cpuType'],
            size['cpuCores'],
            extraDiskSizesGB=[],
            monthlyTrafficPackage=size['monthlyTrafficPackage'],
            id=size['id']
        ) for size in response.object]

    def list_images(self, location):
        """
        List available disk images.

        :param location: Location of the deployement.
                         Available disk images depend on location.
        :type location: :class:`.NodeLocation`

        :rtype: ``list`` of :class:`NodeImage`
        """
        response = self.connection.request(
            'service/server?images=1&datacenter=%s' % location.id)
        images = []
        for image in response.object:
            extra = self._copy_dict(('datacenter', 'os', 'code',
                                     'osDiskSizeGB', 'ramMBMin'), image)
            images.append(self.ex_get_image(image['name'],
                                            image['id'], extra))
        return images

    def create_node(self, name, size, image, location, auth=None,
                    ex_networks=None, ex_dailybackup=False,
                    ex_managed=False, ex_billingcycle=EX_BILLINGCYCLE_HOURLY,
                    ex_poweronaftercreate=True,
                    ex_wait=True):
        """
        Creates a Kamatera node.

        If auth is not given then password will be generated.

        :param name:   String with a name for this new node (required)
        :type name:   ``str``

        :param size:   The size of resources allocated to this node (required)
        :type size:   :class:`.NodeSize`

        :param image:  OS Image to boot on node. (required)
        :type image:  :class:`.NodeImage`

        :param location: Which data center to create a node in. (required)
        :type location: :class:`.NodeLocation`

        :param auth:   Authentication information for the node (optional)
        :type auth:   :class:`.NodeAuthSSHKey` or :class:`.NodeAuthPassword`

        :param ex_networks:   Network configurations (optional)
        :type ex_networks:   ``list`` of ``dict``

        :param ex_dailybackup:   Whether to create daily backups (optional)
        :type ex_dailybackup:    ``bool``

        :param ex_managed:   Whether to provide managed support (optional)
        :type ex_managed:    ``bool``

        :param ex_billingcycle:   billing cycle (hourly / monthly) (optional)
        :type ex_billingcycle:    ``str``

        :param ex_poweronaftercreate:   power on after creation (optional)
        :type ex_poweronaftercreate:    ``bool``

        :param ex_wait:   wait for server to be running (optional)
        :type ex_wait:    ``bool``

        :return: The newly created node.
        :rtype: :class:`.Node`
        """
        password = None
        pubkey = None
        generate_password = False
        if isinstance(auth, basestring):
            password = auth
        else:
            auth_obj = self._get_and_check_auth(auth)
            if getattr(auth_obj, 'generated', False):
                password = '__generate__'
                generate_password = True
            elif hasattr(auth_obj, 'password'):
                password = auth_obj.password
                generate_password = False
            elif hasattr(auth_obj, 'pubkey'):
                pubkey = auth_obj.pubkey
        if not ex_networks:
            ex_networks = [{'name': 'wan', 'ip': 'auto'}]
        request_data = {
            "name": name,
            "password": password or '',
            "passwordValidate": password or '',
            'ssh-key': pubkey or '',
            "datacenter": location.id,
            "image": image.id,
            "cpu": '%s%s' % (size.extra['cpuCores'], size.extra['cpuType']),
            "ram": size.ram,
            "disk": ' '.join([
                'size=%d' % disksize for disksize
                in [size.disk] + size.extra['extraDiskSizesGB']]),
            "dailybackup": 'yes' if ex_dailybackup else 'no',
            "managed": 'yes' if ex_managed else 'no',
            "network": ' '.join([','.join([
                '%s=%s' % (k, v) for k, v
                in network.items()]) for network in ex_networks]),
            "quantity": 1,
            "billingcycle": ex_billingcycle,
            "monthlypackage": size.extra['monthlyTrafficPackage'] or '',
            "poweronaftercreate": 'yes' if ex_poweronaftercreate else 'no'
        }
        response = self.connection.request('service/server', method='POST',
                                           data=json.dumps(request_data))
        if generate_password:
            command_ids = response.object['commandIds']
            generated_password = response.object['password']
        else:
            command_ids = response.object
            generated_password = None
        if len(command_ids) != 1:
            raise RuntimeError('invalid response')
        node = self.ex_get_node(
            name=name, size=size, image=image, location=location,
            dailybackup=ex_dailybackup, managed=ex_managed,
            billingcycle=ex_billingcycle,
            generated_password=generated_password,
            create_command_id=command_ids[0],
            poweronaftercreate=ex_poweronaftercreate)
        if ex_wait:
            if (
                'create_command_id' not in node.extra
                or node.state != NodeState.UNKNOWN
            ):
                raise ValueError('invalid node for updating create status')
            command = self.ex_wait_command(node.extra['create_command_id'])
            node.extra['create_log'] = command.get('log')
            if node.extra.get('poweronaftercreate'):
                node.state = NodeState.RUNNING
            else:
                node.state = NodeState.STOPPED
            if command.get('completed'):
                node.created_at = datetime.datetime.strptime(
                    command['completed'], '%Y-%m-%d %H:%M:%S')
            name_lines = [line for line
                          in node.extra['create_log'].split("\n")
                          if line.startswith('Name: ')]
            if len(name_lines) != 1:
                raise RuntimeError('Invalid node create log response')
            node.name = name_lines[0].replace('Name: ', '')
            response = self.connection.request(
                '/service/server/info', method='POST',
                data=json.dumps({'name': node.name}))
            self._update_node_from_server_info(node, response.object[0])
        return node

    def list_nodes(self, ex_name_regex=None, ex_full_details=False,
                   ex_id=None):
        """
        List nodes

        :param ex_name_regex:   Regular expression to match node names
                                if set returns full node details (optional)
        :type ex_name_regex:    ``str``

        :param ex_full_details:   Whether to return full node details
                                  takes longer to complete (optional)
        :type ex_full_details:    ``bool``

        :return: List of node objects
        :rtype: ``list`` of :class:`Node`
        """
        if ex_name_regex or ex_full_details or ex_id:
            request_data = {}
            if ex_id:
                request_data['id'] = ex_id
            else:
                if not ex_name_regex:
                    ex_name_regex = '.*'
                request_data['name'] = ex_name_regex
            response = self.connection.request(
                '/service/server/info', method='POST',
                data=json.dumps(request_data))
            return [self._update_node_from_server_info(
                self.ex_get_node(), server) for server in response.object]
        else:
            response = self.connection.request('/service/servers')
            return [
                self.ex_get_node(
                    id=server['id'], name=server['name'],
                    state=(
                        NodeState.RUNNING if server['power'] == 'on'
                        else NodeState.STOPPED),
                    location=self.ex_get_location(server['datacenter'])
                ) for server in response.object]

    def reboot_node(self, node, ex_wait=True):
        """
        Reboot the given node

        :param node:     the node to reboot
        :type node: :class:`Node`

        :param ex_wait:     wait for reboot to complete (optional)
        :type ex_wait:  ``bool``

        :rtype: ``bool``
        """
        return self.ex_node_operation(node, 'reboot', ex_wait)

    def destroy_node(self, node, ex_wait=True):
        """
        Destroy the given node

        :param node:     the node to destroy
        :type node: :class:`Node`

        :param ex_wait:     wait for destroy to complete (optional)
        :type ex_wait:      ``bool``

        :rtype: ``bool``
        """
        return self.ex_node_operation(node, 'terminate', ex_wait)

    def stop_node(self, node, ex_wait=True):
        """
        Stop the given node

        :param node:     the node to stop
        :type node: :class:`Node`

        :param ex_wait:     wait for stop to complete (optional)
        :type ex_wait:      ``bool``

        :rtype: ``bool``
        """
        return self.ex_node_operation(node, 'poweroff', ex_wait)

    def start_node(self, node, ex_wait=True):
        """
        Start the given node

        :param node:     the node to start
        :type node: :class:`Node`

        :param ex_wait:     wait for start to complete (optional)
        :type ex_wait:      ``bool``

        :rtype: ``bool``
        """
        return self.ex_node_operation(node, 'poweron', ex_wait)

    def ex_node_operation(self, node, operation, wait=True):
        """
        Run custom operations on the node

        :param node:     the node to run operation on
        :type node: :class:`Node`

        :param operation:   the operation to run
        :type operation:   ``str``

        :param ex_wait:     wait for destroy to complete (optional)
        :type ex_wait:      ``bool``

        :rtype: ``bool``
        """
        if node.id:
            request_data = {'id': node.id}
        elif node.name:
            request_data = {'name': node.name}
        else:
            raise ValueError('Invalid node for %s node operation: '
                             'missing id / name' % operation)
        if operation == 'terminate':
            request_data['force'] = True
        command_id = self.connection.request(
            '/service/server/%s' % operation, method='POST',
            data=json.dumps(request_data)).object[0]
        if wait:
            self.ex_wait_command(command_id)
        else:
            node.extra['%s_command_id' % operation] = command_id
        return True

    def ex_get_location(self, id, name=None, country=None):
        """
        Get a NodeLocation object to use for other methods

        :param id:     Location ID - uppercase letters code (required)
        :type id:      ``str``

        :param name:     Location Name (optional)
        :type name:      ``str``

        :param name:     Location country (optional)
        :type name:      ``str``

        :rtype: :class:`.NodeLocation`
        """
        return NodeLocation(id=id, name=name, country=country, driver=self)

    def ex_get_size(self, ramMB, diskSizeGB, cpuType, cpuCores,
                    extraDiskSizesGB=None, monthlyTrafficPackage=None,
                    id=None, name=None):
        """
        Get a NodeSize object to use for other methods

        :param ramMB:     Amount of RAM to allocate in MB (required)
        :type ramMB:      ``int``

        :param diskSizeGB:     disk size GB for primary hard disk (required)
        :type diskSizeGB:      ``int``

        :param cpuType:     CPU type ID (single uppercase letter),
                            see ex_list_capabilities (required)
        :type cpuType:      ``str``

        :param cpuCores:     Number of CPU cores to allocate (required)
        :type cpuCores:      ``int``

        :param extraDiskSizesGB:     additional disk sizes in GB (optional)
        :type extraDiskSizesGB:      ``list`` of :int:

        :param monthlyTrafficPackage:     ID of monthly traffic package
                                          see ex_list_capabilities (optional)
        :type monthlyTrafficPackage:      ``str``

        :param id:     Size ID (optional)
        :type id:      ``str``

        :param name:     Size Name (optional)
        :type name:      ``str``

        :rtype: :class:`.NodeLocation`
        """
        if not id:
            id = str(cpuCores) + cpuType
            id += '-' + str(ramMB) + 'MB-' + str(diskSizeGB) + 'GB'
            if monthlyTrafficPackage:
                id += '-' + monthlyTrafficPackage
        if not name:
            name = id
        return NodeSize(
            id=id,
            name=name,
            ram=ramMB,
            disk=diskSizeGB,
            bandwidth=0,
            price=0,
            driver=self.connection.driver,
            extra={
                'cpuType': cpuType,
                'cpuCores': cpuCores,
                'monthlyTrafficPackage': monthlyTrafficPackage,
                'extraDiskSizesGB': extraDiskSizesGB or []
            }
        )

    def ex_get_image(self, name=None, id=None, extra=None):
        if not id and not name:
            raise ValueError('either id or name are required for NodeImage')
        return NodeImage(id=id or name,
                         name=name or '', driver=self, extra=extra or {})

    def ex_list_capabilities(self, location):
        """
        List capabilities for given location.

        :param location: Location of the deployment.
        :type location: :class:`.NodeLocation`

        :return: ``dict``
        """
        return self.connection.request(
            'service/server?capabilities=1&'
            'datacenter=%s' % location.id).object

    def ex_wait_command(self, command_id, timeout_seconds=600,
                        poll_interval_seconds=2):
        """
        Wait for command to complete and return the command status details

        :param command_id: Command ID to wait for. (required)
        :type command_id: ``int``

        :param timeout_seconds: Max seconds to wait for command. (optional)
        :type timeout_seconds: ``int``

        :param poll_interval_seconds: Poll interval in seconds (optional)
        :type poll_interval_seconds: ``int``

        :return: ``dict``
        """
        start_time = datetime.datetime.now()
        time.sleep(poll_interval_seconds)
        while True:
            max_time = start_time + datetime.timedelta(
                seconds=timeout_seconds)
            if max_time < datetime.datetime.now():
                raise TimeoutError(
                    'Timeout waiting for command '
                    '(timeout_seconds=%s, command_id=%s)' % (
                        str(timeout_seconds), str(command_id)))
            time.sleep(poll_interval_seconds)
            command = self.ex_get_command_status(command_id)
            status = command.get('status')
            if status == 'complete':
                return command
            elif status == 'error':
                raise RuntimeError('Command failed: ' + command.get('log'))

    def ex_get_command_status(self, command_id):
        """
        Get Kamatera command status details

        :param command_id: Command ID to get details for. (required)
        :type command_id: ``int``

        :return: ``dict``
        """
        response = self.connection.request(
            '/service/queue?id=' + str(command_id))
        if len(response.object) != 1:
            raise RuntimeError('invalid response')
        return response.object[0]

    def ex_get_node(self, id=None, name=None, state=NodeState.UNKNOWN,
                    public_ips=None, private_ips=None, size=None,
                    image=None, created_at=None, location=None,
                    dailybackup=None, managed=None, billingcycle=None,
                    generated_password=None, create_command_id=None,
                    poweronaftercreate=None):
        """
        Get a Kamatera node object.

        :param id:   Node ID (optional)
        :type id:   ``str``

        :param name:   Node name (optional)
        :type name:   ``str``

        :param state:   Node state (optional)
        :type state:   :class:`libcloud.compute.types.NodeState`

        :param public_ips:   Node public IPS. (optional)
        :type public_ips:   ``list`` of :str:

        :param private_ips:   Node private IPS. (optional)
        :type private_ips:   ``list`` of :str:

        :param size:  node size. (optional)
        :type size:  :class:`.NodeSize`

        :param image:  Node OS Image. (optional)
        :type image:  :class:`.NodeImage`

        :param created_at:  Node creation time. (optional)
        :type created_at:  ``datetime.datetime``

        :param location: Node datacenter. (optional)
        :type location: :class:`.NodeLocation`

        :param dailybackup:   create daily backups for the node (optional)
        :type dailybackup:    ``bool``

        :param managed:   provide managed support for the node (optional)
        :type managed:    ``bool``

        :param billingcycle:   billing cycle (hourly / monthly) (optional)
        :type billingcycle:    ``str``

        :param generated_password:   server generated password (optional)
        :type generated_password:    ``str``

        :param create_command_id:   creation task command ID (optional)
        :type create_command_id:    ``int``

        :param poweronaftercreate:   power on the node after create (optional)
        :type poweronaftercreate:    ``bool``

        :return: The node.
        :rtype: :class:`.Node`
        """
        extra = {}
        if location:
            extra['location'] = location
        if dailybackup is not None:
            extra['dailybackup'] = dailybackup
        if managed is not None:
            extra['managed'] = managed
        if billingcycle is not None:
            extra['billingcycle'] = billingcycle
        if generated_password is not None:
            extra['generated_password'] = generated_password
        if create_command_id is not None:
            extra['create_command_id'] = create_command_id
        if poweronaftercreate is not None:
            extra['poweronaftercreate'] = poweronaftercreate
        return Node(id=id, name=name, state=state, public_ips=public_ips,
                    private_ips=private_ips, driver=self,
                    size=size, image=image, created_at=created_at,
                    extra=extra)

    def _copy_dict(self, keys, d):
        extra = {}
        for key in keys:
            extra[key] = d[key]
        return extra

    def _update_node_from_server_info(self, node, server):
        node.id = server['id']
        node.name = server['name']
        if server['power'] == 'on':
            node.state = NodeState.RUNNING
        else:
            node.state = NodeState.STOPPED
        for network in server.get('networks', []):
            if network.get('network').startswith('wan-'):
                node.public_ips += network.get('ips', [])
            else:
                node.private_ips += network.get('ips', [])
        billing = server.get('billing',
                             node.extra.get('billingcycle')).lower()
        if billing == self.EX_BILLINGCYCLE_HOURLY:
            node.extra['billingcycle'] = self.EX_BILLINGCYCLE_HOURLY
            node.extra['priceOn'] = server.get('priceHourlyOn')
            node.extra['priceOff'] = server.get('priceHourlyOff')
        else:
            node.extra['billingcycle'] = self.EX_BILLINGCYCLE_MONTHLY
            node.extra['priceOn'] = server.get('priceMonthlyOn')
            node.extra['priceOff'] = server.get('priceMonthlyOn')
        node.extra['location'] = self.ex_get_location(server['datacenter'])
        node.extra['dailybackup'] = server.get('backup') == "1"
        node.extra['managed'] = server.get('managed') == "1"
        return node



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.compute.drivers.cloudwatt import CloudwattNodeDriver
from libcloud.test.compute.test_openstack import OpenStack_1_1_Tests


class CloudwattNodeDriverTests(OpenStack_1_1_Tests, unittest.TestCase):
    driver_klass = CloudwattNodeDriver
    driver_type = CloudwattNodeDriver

    # These tests dont work because cloudwatt doesn't pass,
    # auth tokens- hide them from the base class
    def test_ex_force_auth_token_passed_to_connection(self):
        pass

    def test_auth_token_without_base_url_raises_exception(self):
        pass

    def test_ex_auth_cache_passed_to_identity_connection(self):
        pass

    def test_unauthorized_clears_cached_auth_context(self):
        pass


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from libcloud.common.gridscale import GridscaleBaseDriver
from libcloud.common.gridscale import GridscaleConnection
from libcloud.compute.base import NodeImage, NodeLocation, VolumeSnapshot, \
    Node, StorageVolume, KeyPair, NodeState, StorageVolumeState, NodeDriver
from libcloud.compute.providers import Provider
from libcloud.utils.iso8601 import parse_date


class GridscaleIp(object):
    """
    Ip Object

    :param id: uuid
    :type id: ``str``
    :param family: family of ip (v4 or v6)
    :type family: ``str``
    :param prefix: prefix of ip
    :type prefix: ``str``
    :param ip_address: Ip address
    :type ip_address: ``str``
    :param create_time: Time ip was created
    :type create_time: ``str``
    """

    def __init__(self, id, family, prefix, create_time, address, extra=None):
        self.id = id
        self.family = family
        self.prefix = prefix
        self.create_time = create_time
        self.ip_address = address
        self.extra = extra or {}

    def __repr__(self):
        return ('Ip: id={}, family={}, prefix={}, create_time={}, '
                'ip_address={}'
                .format(self.id, self.family,
                        self.prefix,
                        self.create_time,
                        self.ip_address))


class GridscaleNetwork(object):
    """
    Network Object

    :param id: uuid
    :type id: ``str``
    :param name: Name of Network
    :type name: ``str``
    :param status: Network status
    :type status: ``str``
    :param relations: object related to network
    :type relations: ``object``
    :param create_time: Time Network was created
    :type create_time: ``str``
    """

    def __init__(self, id, name, status, create_time, relations):
        self.id = id
        self.name = name
        self.status = status
        self.create_time = create_time
        self.relations = relations

    def __repr__(self):
        return ('Network: id={}, name={}, status={}, create_time={}, '
                'relations={}'.format(self.id, self.name, self.status,
                                      self.create_time, self.relations))


class GridscaleNodeDriver(GridscaleBaseDriver, NodeDriver):
    """
    create and entry in libcloud/compute/providers for gridscale
    """
    connectionCls = GridscaleConnection
    type = Provider.GRIDSCALE
    name = 'Gridscale'
    api_name = 'gridscale'
    website = 'https://gridscale.io'
    features = {'create_node': ['ssh_key']}

    def __init__(self, user_id, key, **kwargs):
        super(GridscaleNodeDriver, self).__init__(user_id, key, **kwargs)

    def list_nodes(self):
        """
        List all nodes.

        :return: List of node objects
        :rtype: ``list`` of :class:`.Node`
        """
        result = self._sync_request(data=None, endpoint='objects/servers/')
        nodes = []
        for key, value in self._get_response_dict(result).items():
            node = self._to_node(value)
            nodes.append(node)

        return sorted(nodes, key=lambda sort: sort.created_at)

    def list_locations(self):
        """
        List all available data centers.

        :return: List of node location objects
        :rtype: ``list`` of :class:`.NodeLocation`
        """
        locations = []
        result = self._sync_request(endpoint='objects/locations/')
        for key, value in self._get_response_dict(result).items():
            location = self._to_location(value)
            locations.append(location)

        return sorted(locations, key=lambda nod: nod.id)

    def list_volumes(self):
        """
        List all volumes.

        :return: List of StorageVolume object
        :rtype: ``list`` of :class:`.StorageVolume`
        """
        volumes = []
        result = self._sync_request(endpoint='objects/storages/')
        for key, value in self._get_response_dict(result).items():
            volume = self._to_volume(value)
            volumes.append(volume)
        return sorted(volumes, key=lambda sort: sort.extra['create_time'])

    def ex_list_networks(self):
        """
        List all networks.

        :return: List of objects.
        :rtype: ``list`` of :class:`.GridscaleNetwork`
        """
        networks = []
        result = self._sync_request(endpoint='objects/networks/')
        for key, value in self._get_response_dict(result).items():
            network = self._to_network(value)
            networks.append(network)
        return sorted(networks, key=lambda sort: sort.create_time)

    def list_volume_snapshots(self, volume):
        """
        Lists all snapshots for storage volume.

        :param volume: storage the snapshot is attached to
        :type volume: :class:`.StorageVolume`

        :return: Snapshots
        :rtype: ``list`` of :class:`.VolumeSnapshot`
        """
        snapshots = []
        result = self._sync_request(
            endpoint='objects/storages/'
                     '{}/snapshots'.format(volume.id))
        for key, value in self._get_response_dict(result).items():
            snapshot = self._to_volume_snapshot(value)
            snapshots.append(snapshot)
        return sorted(snapshots, key=lambda snapshot: snapshot.created)

    def ex_list_ips(self):
        """
        Lists all IPs available.

        :return: List of IP objects.
        :rtype: ``list`` of :class:`.GridscaleIp`
        """
        ips = []
        result = self._sync_request(endpoint='objects/ips/')
        for key, value in self._get_response_dict(result).items():
            ip = self._to_ip(value)
            ips.append(ip)
        return ips

    def list_images(self):
        """
        List images.

        :return: List of node image objects
        :rtype: ``list`` of :class:`.NodeImage`
        """
        templates = []
        result = self._sync_request(endpoint='objects/templates')
        for key, value in self._get_response_dict(result).items():
            template = self._to_node_image(value)
            templates.append(template)
        return sorted(templates, key=lambda sort: sort.name)

    def create_node(self, name, size, image, location, ex_ssh_key_ids=None,
                    **kwargs):
        """
        Create a simple node  with a name, cores, memory at the designated
        location.

        :param name: Name of the server.
        :type name: ``str``

        :param size: Nodesize object.
        :type size: :class:`.NodeSize`

        :param image: OS image to attach to the storage.
        :type image: :class:`.GridscaleTemplate`

        :param location: The data center to create a node in.
        :type location: :class:`.NodeLocation`

        :keyword ex_ssh_key_ids: List of SSH key IDs to add to the server.
        :type ex_ssh_key_ids: ``list`` of ``str``

        :return: The newly created Node.
        :rtype: :class:`.Node`

        """

        if size.ram % 1024 != 0:
            raise Exception('Value not accepted. Use a multiple of 1024 e.g.'
                            '1024, 2048, 3072...')
        data = {
            'name': name,
            'cores': size.extra['cores'],
            'memory': int(size.ram / 1024),
            'location_uuid': location.id
        }
        self.connection.async_request('objects/servers/',
                                      data=data,
                                      method='POST')

        node = self._to_node(self._get_resource('servers', self.connection
                                                .poll_response_initial
                                                .object['object_uuid']))

        volume = self._create_volume_from_template(name=image.extra['ostype'],
                                                   size=size.disk,
                                                   location=location,
                                                   template={
                                                   'template_uuid': image.id,
                                                   'sshkeys': ex_ssh_key_ids})

        ip = self.ex_create_ip(4, location, name + '_ip')

        self.attach_volume(node, volume)
        self.ex_link_ip_to_node(node, ip)
        self.ex_link_network_to_node(node, self.ex_list_networks()[0])
        self.ex_start_node(node)

        return self._to_node(self._get_resource('servers', node.id))

    def ex_create_ip(self, family, location, name):
        """
        Create either an ip_v4 ip or a ip_v6.

        :param family: Defines if the ip is v4 or v6 with int 4 or int 6.
        :type family: ``int``

        :param location: Defines which datacenter the created ip
                         responds with.
        :type location: :class:`.NodeLocation`

        :param name: Name of your Ip.
        :type name: ``str``

        :return: Ip
        :rtype: :class:`.GridscaleIp`
        """
        self.connection.async_request(
            'objects/ips/',
            data={
                'name': name,
                'family': family,
                'location_uuid': location.id},
            method='POST')

        return self._to_ip(self._get_resource('ips', self.connection
                                              .poll_response_initial
                                              .object['object_uuid']))

    def ex_create_networks(self, name, location):
        """
        Create a network at the data center location.

        :param name: Name of the network.
        :type name: ``str``

        :param location: Location.
        :type location: :class:`.NodeLocation`

        :return: Network.
        :rtype: :class:`.GridscaleNetwork`
        """
        self.connection.async_request(
            'objects/networks',
            data={
                'name': name,
                'location_uuid': location.id},
            method='POST')

        return self._to_network(self._get_resource('network', self.connection
                                                   .poll_response_initial
                                                   .object['object_uuid']))

    def create_volume(self, size, name, location=None, snapshot=None):
        """
        Create a new volume.

        :param size: Integer in GB.
        :type size: ``int``

        :param name: Name of the volume.
        :type name: ``str``

        :param location: The server location.
        :type location: :class:`.NodeLocation`

        :param snapshot:  Snapshot from which to create the new
                          volume.  (optional)
        :type snapshot: :class:`.VolumeSnapshot`

        :return: Newly created StorageVolume.
        :rtype: :class:`.StorageVolume`
        """

        return self._create_volume_from_template(size, name, location)

    def _create_volume_from_template(self, size, name, location=None,
                                     template=None):
        """
        create Storage

        :param name: name of your Storage unit
        :type name: ``str``

        :param size: Integer in GB.
        :type size: ``int``

        :param location: your server location
        :type location: :class:`.NodeLocation`

        :param template: template to shape the storage capacity to
        :type template: ``dict``

        :return: newly created StorageVolume
        :rtype: :class:`.GridscaleVolumeStorage`
        """
        template = template
        self.connection.async_request(
            'objects/storages/',
            data={
                'name': name,
                'capacity': size,
                'location_uuid': location.id,
                'template': template},
            method='POST')

        return self._to_volume(self._get_resource('storages',
                                                  self.connection
                                                  .poll_response_initial
                                                  .object['object_uuid']))

    def create_volume_snapshot(self, volume, name):
        """
        Creates a snapshot of the current state of your volume,
        you can rollback to.

        :param volume: Volume you want to create a snapshot of.
        :type volume: :class:`.StorageVolume`

        :param name: Name of the snapshot.
        :type name: ``str``

        :return: VolumeSnapshot.
        :rtype: :class:`.VolumeSnapshot`
        """
        self.connection.async_request(
            'objects/storages/{}/snapshots'.format(volume.id),
            data={
                'name': name},
            method='POST')

        return self._to_volume_snapshot(self._get_resource(
            'storages/{}/snapshots'.format(volume.id), self.connection
            .poll_response_initial.object['object_uuid']))

    def create_image(self, node, name):
        """
        Creates an image from a node object.

        :param node: Node to run the task on.
        :type node: :class:`.Node`

        :param name: Name for new image.
        :type name: ``str``

        :return: NodeImage.
        :rtype: :class:`.NodeImage`
        """
        storage_dict = node.extra['relations']['storages'][0]
        snapshot_uuid = ''
        if storage_dict['bootdevice'] is True:
            self.connection.async_request(
                'objects/storages/{}/snapshots/'
                .format(storage_dict['object_uuid']),
                data={'name': name + '_snapshot'},
                method='POST')

            snapshot_uuid = self.connection.poll_response_initial.object[
                'object_uuid']

            self.connection.async_request(
                'objects/templates/',
                data={
                    'name': name,
                    'snapshot_uuid': snapshot_uuid},
                method='POST')

            snapshot_dict = self._get_response_dict(self._sync_request(
                endpoint='objects/storages/{}/snapshots/{}'
                .format(storage_dict['object_uuid'], snapshot_uuid)))

            self.destroy_volume_snapshot(
                self._to_volume_snapshot(snapshot_dict))

        return self._to_node_image(self._get_resource(
            'templates', self.connection.poll_response_initial.object[
                'object_uuid']))

    def destroy_node(self, node, ex_destroy_associated_resources=False):
        """
        Destroy node.

        :param node: Node object.
        :type node: :class:`.Node`

        :param ex_destroy_associated_resources: True to destroy associated
        resources such as storage volumes and IPs.
        :type ex_destroy_associated_resources: ``bool``

        :return: True if the destroy was successful, otherwise False.
        :rtype: ``bool``
        """
        if ex_destroy_associated_resources:
            associated_volumes = self.ex_list_volumes_for_node(node=node)
            associated_ips = self.ex_list_ips_for_node(node=node)

        # 1. Delete the server itself
        result = self._sync_request(endpoint='objects/servers/{}'
                                    .format(node.id),
                                    method='DELETE')

        # 2. Destroy associated resouces (if requested)
        if ex_destroy_associated_resources:
            for volume in associated_volumes:
                self.destroy_volume(volume=volume)

            for ip in associated_ips:
                self.ex_destroy_ip(ip=ip)

        return result.status == 204

    def destroy_volume(self, volume):
        """
        Delete volume.

        :param volume: Volume to be destroyed.
        :type volume: :class:`.StorageVolume`

        :return: True if the destroy was successful, otherwise False.
        :rtype: ``bool``
        """
        result = self._sync_request(endpoint='objects/storages/{}'
                                    .format(volume.id),
                                    method='DELETE')
        return result.status == 204

    def ex_destroy_ip(self, ip):
        """
        Delete an ip.

        :param ip: IP object.
        :type ip: :class:`.GridscaleIp`

        :return: ``True`` if delete_image was successful, ``False`` otherwise.
        :rtype: ``bool``
        """
        result = self._sync_request(endpoint='objects/ips/{}'
                                    .format(ip.id),
                                    method='DELETE')
        return result.status == 204

    def destroy_volume_snapshot(self, snapshot):
        """
        Destroy a snapshot.

        :param snapshot: The snapshot to delete.
        :type snapshot: :class:'.VolumeSnapshot`

        :return: True if the destroy was successful, otherwise False.
        :rtype: ``bool``
        """
        result = self._sync_request(endpoint='objects/storages/'
                                             '{}/snapshots/{}/'
                                    .format(snapshot.extra['parent_uuid'],
                                            snapshot.id),
                                    method='DELETE')
        return result.status == 204

    def ex_destroy_network(self, network):
        """
        Delete network.

        :param network: Network object.
        :type network: :class:`.GridscaleNetwork`

        :return: ``True`` if destroyed successfully, otherwise ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(endpoint='objects/networks/{}'
                                    .format(network.id),
                                    method='DELETE')
        return result.status == 204

    def delete_image(self, node_image):
        """
        Destroy an image.

        :param node_image: Node image object.
        :type node_image: :class:`.NodeImage`

        :return: True if the destroy was successful, otherwise False
        :rtype: ``bool``

        """
        result = self._sync_request(endpoint='objects/templates/{}'
                                    .format(node_image.id),
                                    method='DELETE')
        return result.status == 204

    def ex_rename_node(self, node, name):
        """
        Modify node name.

        :param name: New node name.
        :type name: ``str``

        :param node: Node
        :type node: :class:`.Node`

        :return: ``True`` or ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(data={'name': name},
                                    endpoint='objects/servers/{}'
                                    .format(node.id),
                                    method='PATCH')
        return result.status == 204

    def ex_rename_volume(self, volume, name):
        """
        Modify storage volume name

        :param volume: Storage.
        :type volume: :class:.`StorageVolume`

        :param name: New storage name.
        :type name: ``str``

        :return: ``True`` or ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(data={'name': name},
                                    endpoint='objects/storages/{}'
                                    .format(volume.id),
                                    method='PATCH')
        return result.status == 204

    def ex_rename_network(self, network, name):
        """
        Modify networks name.

        :param network: Network.
        :type network: :class:`.GridscaleNetwork`

        :param name: New network name.
        :type name: ``str``

        :return: ``True`` or ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(data={'name': name},
                                    endpoint='objects/networks/{}'
                                    .format(network.id),
                                    method='PATCH')
        return result.status == 204

    def reboot_node(self, node, ex_sleep_interval=3):
        """
        Reboot a node.

        :param node: Node object.
        :type node: :class:`.Node`

        :return: True if the reboot was successful, otherwise False.
        :rtype: ``bool``

        :keyword ex_sleep_interval: time to let the shutdown process finish
        :type ex_sleep_interval: ``int``

        """

        if node.extra['power'] is True:
            data = dict({'power': False})
            self._sync_request(data=data,
                               endpoint='objects/servers/{}/power'
                               .format(node.id),
                               method='PATCH')
            time.sleep(ex_sleep_interval)

            data = dict({'power': True})
            self._sync_request(data=data,
                               endpoint='objects/servers/{}/power'
                               .format(node.id),
                               method='PATCH')
            return True

        else:

            return False

    def import_key_pair_from_string(self, name, key_material):
        data = {
            'name': name,
            'sshkey': key_material
        }
        result = self._sync_request(endpoint='objects/sshkeys/', method='POST',
                                    data=data)
        key = self._to_key(result.object, name=name, sshkey=key_material)
        return key

    def list_key_pairs(self):
        """
        List all the available key pair objects.

        :rtype: ``list``of :class:`.KeyPair` objects
        """
        keys = []
        result = self._sync_request(endpoint='objects/sshkeys/')
        for key, value in self._get_response_dict(result).items():
            key = self._to_key(value)
            keys.append(key)
        return keys

    def get_image(self, image_id):
        """
        Get an image based on an image_id.

        :param image_id: Image identifier.
        :type image_id: ``str``

        :return: A NodeImage object.
        :rtype: :class:`.NodeImage`
        """

        response_dict = self._get_response_dict(self._sync_request(
            endpoint='/objects/templates/{}'.format(image_id)))

        return self._to_node_image(response_dict)

    def start_node(self, node):
        result = self._sync_request(data={'power': True},
                                    endpoint='objects/servers/{}/power'
                                    .format(node.id),
                                    method='PATCH')

        return result.status == 204

    def ex_start_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0
        return self.start_node(node=node)

    def ex_link_isoimage_to_node(self, node, isoimage):
        """
        link and isoimage to a node

        :param node: Node you want to link the iso image to
        :type node: ``object``

        :param isoimage: isomiage you want to link
        :type isoimage: ``object``

        :return: None -> success
        :rtype: ``None``
        """
        result = self._sync_request(data={'object_uuid': isoimage.id},
                                    endpoint='objects/servers/{}/isoimages/'
                                    .format(node.id),
                                    method='POST')
        return result

    def attach_volume(self, node, volume):
        """
         Attaches volume to node.

        :param node: Node to attach volume to.
        :type node: :class:`.Node`

        :param volume: Volume to attach.
        :type volume: :class:`.StorageVolume`

        :rytpe: ``bool``
        """
        result = self._sync_request(data={'object_uuid': volume.id},
                                    endpoint='objects/servers/{}/storages/'
                                    .format(node.id),
                                    method='POST')
        return result.status == 204

    def ex_link_network_to_node(self, node, network):
        """
        Link a network to a node.

        :param node: Node object to link networks to.
        :type node: :class:`.Node`

        :param network: Network you want to link.
        :type network: :class:`.GridscaleNetwork`

        :return: ``True`` if linked sucessfully, otherwise ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(data={'object_uuid': network.id},
                                    endpoint='objects/servers/{}/networks/'
                                    .format(node.id),
                                    method='POST')
        return result.status == 204

    def ex_link_ip_to_node(self, node, ip):
        """
        links a existing ip with a node

        :param node: node object
        :type node: ``object``

        :param ip: ip object
        :type ip: ``object``

        :return: Request ID
        :rtype: ``str``
        """
        result = self._sync_request(data={'object_uuid': ip.id},
                                    endpoint='objects/servers/{}/ips/'
                                    .format(node.id),
                                    method='POST')
        return result

    def ex_unlink_isoimage_from_node(self, node, isoimage):
        """
        unlink isoimages from server

        :param node: node you want to unlink the image from
        :type node: ``object``

        :param isoimage: isoimage you want to unlink
        :type isoimage: ``object``

        :return: None -> success
        :rtype: ``None``
        """
        result = self._sync_request(endpoint='objects/servers/{}/isoimages/{}'
                                    .format(node.id, isoimage.id),
                                    method='DELETE')
        return result

    def ex_unlink_ip_from_node(self, node, ip):
        """
        unlink ips from server

        :param node: node you want to unlink the ip from
        :type node: ``object``

        :param ip: the ip you want to unlink
        :type ip: ``object``

        :return: None -> success
        :rtype: ``None``
        """
        result = self._sync_request(endpoint='objects/servers/{}/ips/{}'
                                    .format(node.id, ip.id),
                                    method='DELETE')
        return result

    def ex_unlink_network_from_node(self, node, network):
        """
        Unlink network from node.

        :param node: Node you want to unlink from network.
        :type node: :class:`.Node`

        :param network: Network you want to unlink.
        :type network: :class:`.GridscaleNetwork

        :return: ``True`` if unlink was successful, otherwise ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(endpoint='objects/servers/{}/networks/{}'
                                    .format(node.id, network.id),
                                    method='DELETE')
        return result.status == 204

    def detach_volume(self, volume):
        """
        Detaches a volume from a node.

        :param volume: Volume to be detached
        :type volume: :class:`.StorageVolume`

        :rtype: ``bool``
        """
        node = volume.extra['relations']['servers'][0]
        result = self._sync_request(endpoint='objects/servers/{}/storages/{}'
                                    .format(node['object_uuid'], volume.id),
                                    method='DELETE')
        return result.status == 204

    def ex_storage_rollback(self, volume, snapshot, rollback):
        """
        initiate a rollback on your storage

        :param volume: storage uuid
        :type volume: ``string``

        :param snapshot: snapshot uuid
        :type snapshot: ``string``

        :param rollback: variable
        :type rollback: ``bool``

        :return: RequestID
        :rtype: ``str``
        """
        result = self._sync_request(data={'rollback': rollback},
                                    endpoint='objects/storages/{}/snapshots/'
                                             '{}/rollback'
                                    .format(volume.id, snapshot.id),
                                    method='PATCH')
        return result

    def ex_list_volumes_for_node(self, node):
        """
        Return a list of associated volumes for the provided node.

        :rtype: ``list`` of :class:`StorageVolume`
        """
        volumes = self.list_volumes()

        result = []
        for volume in volumes:
            related_servers = volume.extra.get('relations', {}) \
                .get('servers', [])
            for server in related_servers:
                if server['object_uuid'] == node.id:
                    result.append(volume)

        return result

    def ex_list_ips_for_node(self, node):
        """
        Return a list of associated IPs for the provided node.

        :rype: ``list`` of :class:`GridscaleIp`
        """
        ips = self.ex_list_ips()

        result = []
        for ip in ips:
            related_servers = ip.extra.get('relations', {}) \
                .get('servers', [])
            for server in related_servers:
                # TODO: This is not consistent with volumes where key is
                # called "object_uuid"
                if server['server_uuid'] == node.id:
                    result.append(ip)

        return result

    def _to_node(self, data):
        extra_keys = ['cores', 'power', 'memory', 'current_price', 'relations']

        extra = self._extract_values_to_dict(data=data, keys=extra_keys)
        ips = []

        for diction in data['relations']['public_ips']:
            ips.append(diction['ip'])

        state = ''

        if data['power'] is True:
            state = NodeState.RUNNING
        else:
            state = NodeState.STOPPED

        node = Node(id=data['object_uuid'], name=data['name'], state=state,
                    public_ips=ips,
                    created_at=parse_date(data['create_time']),
                    private_ips=None, driver=self.connection.driver,
                    extra=extra)

        return node

    def _to_volume(self, data):
        extra_keys = ['create_time', 'current_price', 'storage_type',
                      'relations']

        extra = self._extract_values_to_dict(data=data, keys=extra_keys)

        storage = StorageVolume(id=data['object_uuid'],
                                name=data['name'], size=data['capacity'],
                                driver=self.connection.driver,
                                extra=extra)

        return storage

    def _to_volume_snapshot(self, data):
        extra_keys = ['labels', 'status', 'usage_in_minutes',
                      'location_country', 'current_price', 'parent_uuid']

        extra = self._extract_values_to_dict(data=data, keys=extra_keys)

        volume_snapshot = VolumeSnapshot(id=data['object_uuid'],
                                         driver=self.connection.driver,
                                         size=data['capacity'], extra=extra,
                                         created=parse_date(
                                             data['create_time']),
                                         state=StorageVolumeState.AVAILABLE,
                                         name=data['name'])

        return volume_snapshot

    def _to_location(self, data):
        location = NodeLocation(id=data['object_uuid'], name=data['name'],
                                country=data['country'],
                                driver=self.connection.driver, )
        return location

    def _to_ip(self, data):
        extra_keys = ['create_time', 'current_price', 'name',
                      'relations', 'reverse_dns', 'status']
        extra = self._extract_values_to_dict(data=data, keys=extra_keys)

        ip = GridscaleIp(id=data['object_uuid'], family=data['family'],
                         prefix=data['prefix'],
                         create_time=data['create_time'],
                         address=data['ip'],
                         extra=extra)

        return ip

    def _to_network(self, data):
        network = GridscaleNetwork(id=data['object_uuid'], name=data['name'],
                                   create_time=data['create_time'],
                                   status=data['status'],
                                   relations=data['relations'])
        return network

    def _to_node_image(self, data):
        extra_keys = ['capacity', 'create_time', 'labels', 'ostype',
                      'location_name', 'private', 'status',
                      'usage_in_minutes', 'version']

        extra = self._extract_values_to_dict(data=data, keys=extra_keys)

        template = NodeImage(id=data['object_uuid'], name=data['name'],
                             driver=self.connection.driver,
                             extra=extra)

        return template

    def _to_key(self, data, name=None, sshkey=None):
        extra = {
            'uuid': data['object_uuid'],
            'labels': data.get('labels', [])
        }

        name = data.get('name', name)
        sshkey = data.get('sshkey', sshkey)

        key = KeyPair(name=name, fingerprint=data['object_uuid'],
                      public_key=sshkey, private_key=None, extra=extra,
                      driver=self.connection.driver)

        return key

    def _extract_values_to_dict(self, data, keys):
        """
        Extract extra values to dict.

        :param data: dict to extract values from.
        :type data: ``dict``
        :param keys: keys to extract
        :type keys: ``List``
        :return: dictionary containing extra values
        :rtype: ``dict``
        """

        result = {}

        for key in keys:
            if key == 'memory':
                result[key] = data[key] * 1024
            else:
                result[key] = data[key]

        return result

    def _get_response_dict(self, raw_response):
        """

        Get the actual response dictionary.

        :param raw_response: Nested dictionary.
        :type raw_response: ``dict``

        :return: Not-nested dictionary.
        :rtype: ``dict``
        """
        return list(raw_response.object.values())[0]

    def _get_resource(self, endpoint_suffix, object_uuid):
        """
        Get specific uuid specific resource.

        :param endpoint_suffix: Endpoint resource e.g. servers/.
        :type endpoint_suffix: ``str``
        :param object_uuid: Uuid of resource to be pulled.
        :type object_uuid: ``str``
        :return: Response dictionary.
        :rtype: nested ``dict``
        """
        data = self._sync_request(
            endpoint='objects/{}/{}'.format(endpoint_suffix, object_uuid))

        data = self._get_response_dict(data)

        return data



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import time

from libcloud.common.exceptions import BaseHTTPError
from libcloud.common.types import LibcloudError


class UpcloudTimeoutException(LibcloudError):
    pass


class UpcloudCreateNodeRequestBody(object):
    """
    Body of the create_node request

    Takes the create_node arguments (**kwargs) and constructs the request body

    :param      name: Name of the created server (required)
    :type       name: ``str``

    :param      size: The size of resources allocated to this node.
    :type       size: :class:`.NodeSize`

    :param      image: OS Image to boot on node.
    :type       image: :class:`.NodeImage`

    :param      location: Which data center to create a node in. If empty,
                        undefined behavior will be selected. (optional)
    :type       location: :class:`.NodeLocation`

    :param      auth: Initial authentication information for the node
                            (optional)
    :type       auth: :class:`.NodeAuthSSHKey`

    :param      ex_hostname: Hostname. Default is 'localhost'. (optional)
    :type       ex_hostname: ``str``

    :param ex_username: User's username, which is created.
                        Default is 'root'. (optional)
    :type ex_username: ``str``
    """

    def __init__(self, name, size, image, location, auth=None,
                 ex_hostname='localhost', ex_username='root'):
        self.body = {
            'server': {
                'title': name,
                'hostname': ex_hostname,
                'plan': size.id,
                'zone': location.id,
                'login_user': _LoginUser(ex_username, auth).to_dict(),
                'storage_devices': _StorageDevice(image, size).to_dict()
            }
        }

    def to_json(self):
        """
        Serializes the body to json

        :return: JSON string
        :rtype: ``str``
        """
        return json.dumps(self.body)


class UpcloudNodeDestroyer(object):
    """
    Helper class for destroying node.
    Node must be first stopped and then it can be
    destroyed

    :param  upcloud_node_operations: UpcloudNodeOperations instance
    :type   upcloud_node_operations: :class:`.UpcloudNodeOperations`

    :param  sleep_func: Callable function, which sleeps.
        Takes int argument to sleep in seconds (optional)
    :type   sleep_func: ``function``

    """

    WAIT_AMOUNT = 2
    SLEEP_COUNT_TO_TIMEOUT = 20

    def __init__(self, upcloud_node_operations, sleep_func=None):
        self._operations = upcloud_node_operations
        self._sleep_func = sleep_func or time.sleep
        self._sleep_count = 0

    def destroy_node(self, node_id):
        """
        Destroys the given node.

        :param  node_id: Id of the Node.
        :type   node_id: ``int``
        """
        self._stop_called = False
        self._sleep_count = 0
        return self._do_destroy_node(node_id)

    def _do_destroy_node(self, node_id):
        state = self._operations.get_node_state(node_id)
        if state == 'stopped':
            self._operations.destroy_node(node_id)
            return True
        elif state == 'error':
            return False
        elif state == 'started':
            if not self._stop_called:
                self._operations.stop_node(node_id)
                self._stop_called = True
            else:
                # Waiting for started state to change and
                # not calling stop again
                self._sleep()
            return self._do_destroy_node(node_id)
        elif state == 'maintenance':
            # Lets wait maintenace state to go away and retry destroy
            self._sleep()
            return self._do_destroy_node(node_id)
        elif state is None:  # Server not found any more
            return True

    def _sleep(self):
        if self._sleep_count > self.SLEEP_COUNT_TO_TIMEOUT:
            raise UpcloudTimeoutException("Timeout, could not destroy node")
        self._sleep_count += 1
        self._sleep_func(self.WAIT_AMOUNT)


class UpcloudNodeOperations(object):
    """
    Helper class to start and stop node.

    :param  conneciton: Connection instance
    :type   connection: :class:`.UpcloudConnection`
    """

    def __init__(self, connection):
        self.connection = connection

    def stop_node(self, node_id):
        """
        Stops the node

        :param  node_id: Id of the Node
        :type   node_id: ``int``
        """
        body = {
            'stop_server': {
                'stop_type': 'hard'
            }
        }
        self.connection.request('1.2/server/{0}/stop'.format(node_id),
                                method='POST',
                                data=json.dumps(body))

    def get_node_state(self, node_id):
        """
        Get the state of the node.

        :param  node_id: Id of the Node
        :type   node_id: ``int``

        :rtype: ``str``
        """

        action = '1.2/server/{0}'.format(node_id)
        try:
            response = self.connection.request(action)
            return response.object['server']['state']
        except BaseHTTPError as e:
            if e.code == 404:
                return None
            raise

    def destroy_node(self, node_id):
        """
        Destroys the node.

        :param  node_id: Id of the Node
        :type   node_id: ``int``
        """
        self.connection.request('1.2/server/{0}'.format(node_id),
                                method='DELETE')


class PlanPrice(object):
    """
    Helper class to construct plan price in different zones

    :param  zone_prices: List of prices in different zones in UpCloud
    :type   zone_prices: ```list```

    """

    def __init__(self, zone_prices):
        self._zone_prices = zone_prices

    def get_price(self, plan_name, location=None):
        """
        Returns the plan's price in location. If location
        is not provided returns None

        :param  plan_name: Name of the plan
        :type   plan_name: ```str```

        :param  location: Location, which price is returned (optional)
        :type   location: :class:`.NodeLocation`


        rtype: ``float``
        """
        if location is None:
            return None
        server_plan_name = 'server_plan_' + plan_name

        for zone_price in self._zone_prices:
            if zone_price['name'] == location.id:
                return zone_price.get(server_plan_name, {}).get('price')
        return None


class _LoginUser(object):

    def __init__(self, user_id, auth=None):
        self.user_id = user_id
        self.auth = auth

    def to_dict(self):
        login_user = {'username': self.user_id}
        if self.auth is not None:
            login_user['ssh_keys'] = {
                'ssh_key': [self.auth.pubkey]
            }
        else:
            login_user['create_password'] = 'yes'

        return login_user


class _StorageDevice(object):

    def __init__(self, image, size):
        self.image = image
        self.size = size

    def to_dict(self):
        extra = self.image.extra
        if extra['type'] == 'template':
            return self._storage_device_for_template_image()
        elif extra['type'] == 'cdrom':
            return self._storage_device_for_cdrom_image()

    def _storage_device_for_template_image(self):
        hdd_device = {
            'action': 'clone',
            'storage': self.image.id
        }
        hdd_device.update(self._common_hdd_device())
        return {'storage_device': [hdd_device]}

    def _storage_device_for_cdrom_image(self):
        hdd_device = {'action': 'create'}
        hdd_device.update(self._common_hdd_device())
        storage_devices = {
            'storage_device': [
                hdd_device,
                {
                    'action': 'attach',
                    'storage': self.image.id,
                    'type': 'cdrom'
                }
            ]
        }
        return storage_devices

    def _common_hdd_device(self):
        return {
            'title': self.image.name,
            'size': self.size.disk,
            'tier': self.size.extra.get('storage_tier', 'maxiops')
        }



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.loadbalancer.types import Provider
from libcloud.loadbalancer.types import OLD_CONSTANT_TO_NEW_MAPPING
from libcloud.common.providers import get_driver as _get_provider_driver
from libcloud.common.providers import set_driver as _set_provider_driver

__all__ = [
    "Provider",
    "DRIVERS",
    "get_driver",
]

DRIVERS = {
    Provider.RACKSPACE:
    ('libcloud.loadbalancer.drivers.rackspace', 'RackspaceLBDriver'),
    Provider.GOGRID:
    ('libcloud.loadbalancer.drivers.gogrid', 'GoGridLBDriver'),
    Provider.NINEFOLD:
    ('libcloud.loadbalancer.drivers.ninefold', 'NinefoldLBDriver'),
    Provider.BRIGHTBOX:
    ('libcloud.loadbalancer.drivers.brightbox', 'BrightboxLBDriver'),
    Provider.ELB:
    ('libcloud.loadbalancer.drivers.elb', 'ElasticLBDriver'),
    Provider.ALB:
        ('libcloud.loadbalancer.drivers.alb', 'ApplicationLBDriver'),
    Provider.CLOUDSTACK:
    ('libcloud.loadbalancer.drivers.cloudstack', 'CloudStackLBDriver'),
    Provider.GCE:
    ('libcloud.loadbalancer.drivers.gce', 'GCELBDriver'),
    Provider.SOFTLAYER:
    ('libcloud.loadbalancer.drivers.softlayer', 'SoftlayerLBDriver'),
    Provider.DIMENSIONDATA:
    ('libcloud.loadbalancer.drivers.dimensiondata', 'DimensionDataLBDriver'),
    Provider.ALIYUN_SLB:
    ('libcloud.loadbalancer.drivers.slb', 'SLBDriver'),
    Provider.NTTCIS:
    ('libcloud.loadbalancer.drivers.nttcis', 'NttCisLBDriver'),
}


def get_driver(provider):
    deprecated_constants = OLD_CONSTANT_TO_NEW_MAPPING
    return _get_provider_driver(drivers=DRIVERS, provider=provider,
                                deprecated_constants=deprecated_constants)


def set_driver(provider, module, klass):
    return _set_provider_driver(drivers=DRIVERS, provider=provider,
                                module=module, klass=klass)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import sys
import pytest

from libcloud.common.types import InvalidCredsError

from libcloud.utils.py3 import httplib
from libcloud.utils.py3 import xmlrpclib
from libcloud.utils.py3 import next

from libcloud.compute.drivers.softlayer import SoftLayerNodeDriver as SoftLayer
from libcloud.compute.drivers.softlayer import SoftLayerException, \
    NODE_STATE_MAP
from libcloud.compute.types import NodeState, KeyPairDoesNotExistError

from libcloud.test import unittest
from libcloud.test import MockHttp               # pylint: disable-msg=E0611
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.test.secrets import SOFTLAYER_PARAMS

null_fingerprint = '00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:' + \
                   '00:00:00:00:00'


class SoftLayerTests(unittest.TestCase):

    def setUp(self):
        SoftLayer.connectionCls.conn_class = SoftLayerMockHttp
        SoftLayerMockHttp.type = None
        self.driver = SoftLayer(*SOFTLAYER_PARAMS)

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        node = nodes[0]
        self.assertEqual(node.name, 'libcloud-testing1.example.com')
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertEqual(node.extra['password'], 'L3TJVubf')

    def test_initializing_state(self):
        nodes = self.driver.list_nodes()
        node = nodes[1]
        self.assertEqual(node.state, NODE_STATE_MAP['INITIATING'])

    def test_list_locations(self):
        locations = self.driver.list_locations()
        dal = next(l for l in locations if l.id == 'dal05')
        self.assertEqual(dal.country, 'US')
        self.assertEqual(dal.id, 'dal05')
        self.assertEqual(dal.name, 'Dallas - Central U.S.')

    def test_list_images(self):
        images = self.driver.list_images()
        image = images[0]
        self.assertEqual(image.id, 'CENTOS_6_64')

    def test_get_image(self):
        image = self.driver.get_image('CENTOS_6_64')
        self.assertEqual(image.id, 'CENTOS_6_64')

    def test_fail_get_image(self):
        self.assertRaises(
            SoftLayerException,
            self.driver.get_image,
            'NOT_IMAGE')

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()
        self.assertEqual(len(sizes), 13)
        for size in sizes:
            self.assertTrue(size.price > 0.0)

    def test_create_node(self):
        node = self.driver.create_node(name="libcloud-testing",
                                       location=self.driver.list_locations()[0],
                                       size=self.driver.list_sizes()[0],
                                       image=self.driver.list_images()[0])
        self.assertEqual(node.name, 'libcloud-testing.example.com')
        self.assertEqual(node.state, NODE_STATE_MAP['RUNNING'])

    def test_create_node_ex_hourly_True(self):
        SoftLayerMockHttp.type = 'HOURLY_BILLING_TRUE'

        node = self.driver.create_node(name="libcloud-testing",
                                       location=self.driver.list_locations()[0],
                                       size=self.driver.list_sizes()[0],
                                       image=self.driver.list_images()[0],
                                       ex_hourly=True)
        self.assertEqual(node.name, 'libcloud-testing.example.com')
        self.assertEqual(node.state, NODE_STATE_MAP['RUNNING'])

        SoftLayerMockHttp.type = 'HOURLY_BILLING_FALSE'

        node = self.driver.create_node(name="libcloud-testing",
                                       location=self.driver.list_locations()[0],
                                       size=self.driver.list_sizes()[0],
                                       image=self.driver.list_images()[0],
                                       ex_hourly=False)
        self.assertEqual(node.name, 'libcloud-testing.example.com')
        self.assertEqual(node.state, NODE_STATE_MAP['RUNNING'])


    def test_create_fail(self):
        SoftLayerMockHttp.type = "SOFTLAYEREXCEPTION"
        self.assertRaises(
            SoftLayerException,
            self.driver.create_node,
            name="SOFTLAYEREXCEPTION",
            location=self.driver.list_locations()[0],
            size=self.driver.list_sizes()[0],
            image=self.driver.list_images()[0])

    def test_create_creds_error(self):
        SoftLayerMockHttp.type = "INVALIDCREDSERROR"
        self.assertRaises(
            InvalidCredsError,
            self.driver.create_node,
            name="INVALIDCREDSERROR",
            location=self.driver.list_locations()[0],
            size=self.driver.list_sizes()[0],
            image=self.driver.list_images()[0])

    def test_create_node_no_location(self):
        self.driver.create_node(name="Test",
                                size=self.driver.list_sizes()[0],
                                image=self.driver.list_images()[0])

    def test_create_node_no_image(self):
        self.driver.create_node(name="Test", size=self.driver.list_sizes()[0])

    def test_create_node_san(self):
        self.driver.create_node(name="Test", ex_local_disk=False)

    def test_create_node_domain_for_name(self):
        self.driver.create_node(name="libcloud.org")

    def test_create_node_ex_options(self):
        self.driver.create_node(name="Test",
                                location=self.driver.list_locations()[0],
                                size=self.driver.list_sizes()[0],
                                image=self.driver.list_images()[0],
                                ex_domain='libcloud.org',
                                ex_cpus=2,
                                ex_ram=2048,
                                ex_disk=100,
                                ex_keyname='test1',
                                ex_bandwidth=10,
                                ex_local_disk=False,
                                ex_datacenter='Dal05',
                                ex_os='UBUNTU_LATEST')

    def test_reboot_node(self):
        node = self.driver.list_nodes()[0]
        self.driver.reboot_node(node)

    def test_destroy_node(self):
        node = self.driver.list_nodes()[0]
        self.driver.destroy_node(node)

    def test_list_keypairs(self):
        keypairs = self.driver.list_key_pairs()
        self.assertEqual(len(keypairs), 2)
        self.assertEqual(keypairs[0].name, 'test1')
        self.assertEqual(keypairs[0].fingerprint, null_fingerprint)

    def test_get_key_pair(self):
        key_pair = self.driver.get_key_pair(name='test1')
        self.assertEqual(key_pair.name, 'test1')

    def test_get_key_pair_does_not_exist(self):
        self.assertRaises(KeyPairDoesNotExistError, self.driver.get_key_pair,
                          name='test-key-pair')

    @pytest.mark.skip(reason="no way of currently testing this")
    def test_create_key_pair(self):
        key_pair = self.driver.create_key_pair(name='my-key-pair')
        fingerprint = ('1f:51:ae:28:bf:89:e9:d8:1f:25:5d'
                       ':37:2d:7d:b8:ca:9f:f5:f1:6f')

        self.assertEqual(key_pair.name, 'my-key-pair')
        self.assertEqual(key_pair.fingerprint, fingerprint)
        self.assertTrue(key_pair.private_key is not None)

    def test_delete_key_pair(self):
        success = self.driver.delete_key_pair('test1')
        self.assertTrue(success)


class SoftLayerMockHttp(MockHttp, unittest.TestCase):
    fixtures = ComputeFileFixtures('softlayer')

    def _get_method_name(self, type, use_param, qs, path):
        return "_xmlrpc"

    def _xmlrpc(self, method, url, body, headers):
        params, meth_name = xmlrpclib.loads(body)
        url = url.replace("/", "_")
        meth_name = "%s_%s" % (url, meth_name)
        return getattr(self, meth_name)(method, url, body, headers)

    def _xmlrpc_v3_SoftLayer_Virtual_Guest_getCreateObjectOptions(
            self, method, url, body, headers):
        body = self.fixtures.load(
            'v3__SoftLayer_Virtual_Guest_getCreateObjectOptions.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Account_getVirtualGuests(
            self, method, url, body, headers):
        body = self.fixtures.load('v3_SoftLayer_Account_getVirtualGuests.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Location_Datacenter_getDatacenters(
            self, method, url, body, headers):
        body = self.fixtures.load(
            'v3_SoftLayer_Location_Datacenter_getDatacenters.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Virtual_Guest_createObject(
            self, method, url, body, headers):
        fixture = {
            None: 'v3__SoftLayer_Virtual_Guest_createObject.xml',
            'INVALIDCREDSERROR': 'SoftLayer_Account.xml',
            'SOFTLAYEREXCEPTION': 'fail.xml',
            'HOURLY_BILLING_TRUE': 'v3__SoftLayer_Virtual_Guest_createObject.xml',
            'HOURLY_BILLING_FALSE': 'v3__SoftLayer_Virtual_Guest_createObject.xml',
        }[self.type]

        if self.type == 'HOURLY_BILLING_TRUE':
            # Verify parameter is sent as a boolean and not as a string
            expected_value = """
<member>
<name>hourlyBillingFlag</name>
<value><boolean>1</boolean></value>
</member>
""".strip()

            self.assertTrue(expected_value in body, 'Request body is missing hourlyBillingFlag attribute')
        elif self.type == 'HOURLY_BILLING_FALSE':
            # Verify parameter is sent as a boolean and not as a string
            expected_value = """
<member>
<name>hourlyBillingFlag</name>
<value><boolean>0</boolean></value>
</member>
""".strip()

            self.assertTrue(expected_value in body, 'Request body is missing hourlyBillingFlag attribute')


        body = self.fixtures.load(fixture)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Virtual_Guest_getObject(
            self, method, url, body, headers):
        body = self.fixtures.load(
            'v3__SoftLayer_Virtual_Guest_getObject.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Virtual_Guest_rebootSoft(
            self, method, url, body, headers):
        body = self.fixtures.load('empty.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Virtual_Guest_deleteObject(
            self, method, url, body, headers):
        body = self.fixtures.load('empty.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Account_getSshKeys(
            self, method, url, body, headers):
        body = self.fixtures.load('v3__SoftLayer_Account_getSshKeys.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Security_Ssh_Key_getObject(
            self, method, url, body, headers):
        body = self.fixtures.load('v3__SoftLayer_Security_Ssh_Key_getObject.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Security_Ssh_Key_createObject(
            self, method, url, body, headers):
        body = self.fixtures.load('v3__SoftLayer_Security_Ssh_Key_createObject.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _xmlrpc_v3_SoftLayer_Security_Ssh_Key_deleteObject(
            self, method, url, body, headers):
        body = self.fixtures.load('v3__SoftLayer_Security_Ssh_Key_deleteObject.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.base import ConnectionKey, JsonResponse
import json
import base64
import requests
import os
import time

# normally we only use itsyou.online but you might want to work
# against staging.itsyou.online for some testing
IYO_URL = os.environ.get("IYO_URL", "https://itsyou.online")


class G8Connection(ConnectionKey):
    """
    Connection class for G8
    """
    responseCls = JsonResponse

    def add_default_headers(self, headers):
        """
        Add headers that are necessary for every request
        """
        self.driver.key = maybe_update_jwt(self.driver.key)
        headers['Authorization'] = "bearer {}".format(self.driver.key)
        headers['Content-Type'] = 'application/json'
        return headers


def base64url_decode(input):
    """
    Helper method to base64url_decode a string.

    :param input: Input to decode
    :type  input: str

    :rtype: str
    """
    rem = len(input) % 4

    if rem > 0:
        input += b'=' * (4 - rem)
    return base64.urlsafe_b64decode(input)


def is_jwt_expired(jwt):
    """
    Check if jwt is expired

    :param jwt: jwt token to validate expiration
    :type  jwt: str

    :rtype: bool
    """
    jwt = jwt.encode('utf-8')
    signing_input, _ = jwt.rsplit(b'.', 1)
    _, claims_segment = signing_input.split(b'.', 1)
    claimsdata = base64url_decode(claims_segment)
    if isinstance(claimsdata, bytes):
        claimsdata = claimsdata.decode('utf-8')
    data = json.loads(claimsdata)
    # check if it's about to expire in the next minute
    return data['exp'] < time.time() + 60


def maybe_update_jwt(jwt):
    """
    Update jwt if it is expired

    :param jwt: jwt token to validate expiration
    :type  jwt: str

    :rtype: str
    """

    if is_jwt_expired(jwt):
        return refresh_jwt(jwt)
    return jwt


def refresh_jwt(jwt):
    """
    Refresh jwt

    :param jwt: jwt token to refresh
    :type  jwt: str

    :rtype: str
    """
    url = IYO_URL + "/v1/oauth/jwt/refresh"
    headers = {"Authorization": "bearer {}".format(jwt)}
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    return response.text



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Enomaly ECP driver
"""
import time
import os
import socket
import binascii

from libcloud.utils.py3 import httplib
from libcloud.utils.py3 import b
from libcloud.utils.py3 import base64_encode_string

# JSON is included in the standard library starting with Python 2.6.  For 2.5
# and 2.4, there's a simplejson egg at: http://pypi.python.org/pypi/simplejson
try:
    import simplejson as json
except ImportError:
    import json

from libcloud.common.base import Response, ConnectionUserAndKey
from libcloud.compute.base import NodeDriver, NodeSize, NodeLocation
from libcloud.compute.base import NodeImage, Node
from libcloud.compute.types import Provider, NodeState, InvalidCredsError
from libcloud.utils.networking import is_private_subnet

# Defaults
API_HOST = ''
API_PORT = (80, 443)


class ECPResponse(Response):
    def success(self):
        if self.status == httplib.OK or self.status == httplib.CREATED:
            try:
                j_body = json.loads(self.body)
            except ValueError:
                self.error = "JSON response cannot be decoded."
                return False
            if j_body['errno'] == 0:
                return True
            else:
                self.error = "ECP error: %s" % j_body['message']
                return False
        elif self.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError()
        else:
            self.error = "HTTP Error Code: %s" % self.status
        return False

    def parse_error(self):
        return self.error

    # Interpret the json responses - no error checking required
    def parse_body(self):
        return json.loads(self.body)

    def getheaders(self):
        return self.headers


class ECPConnection(ConnectionUserAndKey):
    """
    Connection class for the Enomaly ECP driver
    """

    responseCls = ECPResponse
    host = API_HOST
    port = API_PORT

    def add_default_headers(self, headers):
        # Authentication
        username = self.user_id
        password = self.key
        base64string = base64_encode_string(
            b('%s:%s' % (username, password)))[:-1]
        authheader = "Basic %s" % base64string
        headers['Authorization'] = authheader

        return headers

    def _encode_multipart_formdata(self, fields):
        """
        Based on Wade Leftwich's function:
        http://code.activestate.com/recipes/146306/
        """
        # use a random boundary that does not appear in the fields
        boundary = ''
        while boundary in ''.join(fields):
            boundary = binascii.hexlify(os.urandom(16)).decode('utf-8')
        L = []
        for i in fields:
            L.append('--' + boundary)
            L.append('Content-Disposition: form-data; name="%s"' % i)
            L.append('')
            L.append(fields[i])
        L.append('--' + boundary + '--')
        L.append('')
        body = '\r\n'.join(L)
        content_type = 'multipart/form-data; boundary=%s' % boundary
        header = {'Content-Type': content_type}
        return header, body


class ECPNodeDriver(NodeDriver):
    """
    Enomaly ECP node driver
    """

    name = "Enomaly Elastic Computing Platform"
    website = 'http://www.enomaly.com/'
    type = Provider.ECP
    connectionCls = ECPConnection

    def list_nodes(self):
        """
        Returns a list of all running Nodes

        :rtype: ``list`` of :class:`Node`
        """

        # Make the call
        res = self.connection.request('/rest/hosting/vm/list').parse_body()

        # Put together a list of node objects
        nodes = []
        for vm in res['vms']:
            node = self._to_node(vm)
            if node is not None:
                nodes.append(node)

        # And return it
        return nodes

    def _to_node(self, vm):
        """
        Turns a (json) dictionary into a Node object.
        This returns only running VMs.
        """

        # Check state
        if not vm['state'] == "running":
            return None

        # IPs
        iplist = [interface['ip'] for interface in vm['interfaces'] if
                  interface['ip'] != '127.0.0.1']

        public_ips = []
        private_ips = []
        for ip in iplist:
            try:
                socket.inet_aton(ip)
            except socket.error:
                # not a valid ip
                continue
            if is_private_subnet(ip):
                private_ips.append(ip)
            else:
                public_ips.append(ip)

        # Create the node object
        n = Node(
            id=vm['uuid'],
            name=vm['name'],
            state=NodeState.RUNNING,
            public_ips=public_ips,
            private_ips=private_ips,
            driver=self,
        )

        return n

    def reboot_node(self, node):
        """
        Shuts down a VM and then starts it again.

        @inherits: :class:`NodeDriver.reboot_node`
        """

        # Turn the VM off
        # Black magic to make the POST requests work
        d = self.connection._encode_multipart_formdata({'action': 'stop'})
        self.connection.request(
            '/rest/hosting/vm/%s' % node.id,
            method='POST',
            headers=d[0],
            data=d[1]
        ).parse_body()

        node.state = NodeState.REBOOTING
        # Wait for it to turn off and then continue (to turn it on again)
        while node.state == NodeState.REBOOTING:
            # Check if it's off.
            response = self.connection.request(
                '/rest/hosting/vm/%s' % node.id
            ).parse_body()
            if response['vm']['state'] == 'off':
                node.state = NodeState.TERMINATED
            else:
                time.sleep(5)

        # Turn the VM back on.
        # Black magic to make the POST requests work
        d = self.connection._encode_multipart_formdata({'action': 'start'})
        self.connection.request(
            '/rest/hosting/vm/%s' % node.id,
            method='POST',
            headers=d[0],
            data=d[1]
        ).parse_body()

        node.state = NodeState.RUNNING
        return True

    def destroy_node(self, node):
        """
        Shuts down and deletes a VM.

        @inherits: :class:`NodeDriver.destroy_node`
        """

        # Shut down first
        # Black magic to make the POST requests work
        d = self.connection._encode_multipart_formdata({'action': 'stop'})
        self.connection.request(
            '/rest/hosting/vm/%s' % node.id,
            method='POST',
            headers=d[0],
            data=d[1]
        ).parse_body()

        # Ensure there was no application level error
        node.state = NodeState.PENDING
        # Wait for the VM to turn off before continuing
        while node.state == NodeState.PENDING:
            # Check if it's off.
            response = self.connection.request(
                '/rest/hosting/vm/%s' % node.id
            ).parse_body()
            if response['vm']['state'] == 'off':
                node.state = NodeState.TERMINATED
            else:
                time.sleep(5)

        # Delete the VM
        # Black magic to make the POST requests work
        d = self.connection._encode_multipart_formdata({'action': 'delete'})
        self.connection.request(
            '/rest/hosting/vm/%s' % (node.id),
            method='POST',
            headers=d[0],
            data=d[1]
        ).parse_body()

        return True

    def list_images(self, location=None):
        """
        Returns a list of all package templates aka appliances aka images.

        @inherits: :class:`NodeDriver.list_images`
        """

        # Make the call
        response = self.connection.request(
            '/rest/hosting/ptemplate/list').parse_body()

        # Turn the response into an array of NodeImage objects
        images = []
        for ptemplate in response['packages']:
            images.append(NodeImage(
                id=ptemplate['uuid'],
                name='%s: %s' % (ptemplate['name'], ptemplate['description']),
                driver=self,)
            )

        return images

    def list_sizes(self, location=None):
        """
        Returns a list of all hardware templates

        @inherits: :class:`NodeDriver.list_sizes`
        """

        # Make the call
        response = self.connection.request(
            '/rest/hosting/htemplate/list').parse_body()

        # Turn the response into an array of NodeSize objects
        sizes = []
        for htemplate in response['templates']:
            sizes.append(NodeSize(
                id=htemplate['uuid'],
                name=htemplate['name'],
                ram=htemplate['memory'],
                disk=0,  # Disk is independent of hardware template.
                bandwidth=0,  # There is no way to keep track of bandwidth.
                price=0,  # The billing system is external.
                driver=self,)
            )

        return sizes

    def list_locations(self):
        """
        This feature does not exist in ECP. Returns hard coded dummy location.

        :rtype: ``list`` of :class:`NodeLocation`
        """
        return [NodeLocation(id=1,
                             name="Cloud",
                             country='',
                             driver=self),
                ]

    def create_node(self, name, size, image):
        """
        Creates a virtual machine.

        :keyword    name:   String with a name for this new node (required)
        :type       name:   ``str``

        :keyword    size:   The size of resources allocated to this node .
                            (required)
        :type       size:   :class:`NodeSize`

        :keyword    image:  OS Image to boot on node. (required)
        :type       image:  :class:`NodeImage`

        :rtype: :class:`Node`
        """

        # Find out what network to put the VM on.
        res = self.connection.request(
            '/rest/hosting/network/list').parse_body()

        # Use the first / default network because there is no way to specific
        # which one
        network = res['networks'][0]['uuid']

        # Prepare to make the VM
        data = {
            'name': str(name),
            'package': str(image.id),
            'hardware': str(size.id),
            'network_uuid': str(network),
            'disk': ''
        }

        # Black magic to make the POST requests work
        d = self.connection._encode_multipart_formdata(data)
        response = self.connection.request(
            '/rest/hosting/vm/',
            method='PUT',
            headers=d[0],
            data=d[1]
        ).parse_body()

        # Create a node object and return it.
        n = Node(
            id=response['machine_id'],
            name=data['name'],
            state=NodeState.PENDING,
            public_ips=[],
            private_ips=[],
            driver=self,
        )

        return n



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.common.types import LibcloudError

from libcloud.storage.base import Container, Object
from libcloud.storage.drivers.digitalocean_spaces import (
    DigitalOceanSpacesStorageDriver,
    DOSpacesConnectionAWS4,
    DOSpacesConnectionAWS2,
    DO_SPACES_HOSTS_BY_REGION)

from libcloud.test import LibcloudTestCase
from libcloud.test.secrets import STORAGE_S3_PARAMS


class DigitalOceanSpacesTests(LibcloudTestCase):
    driver_type = DigitalOceanSpacesStorageDriver
    driver_args = STORAGE_S3_PARAMS
    default_host = 'nyc3.digitaloceanspaces.com'

    @classmethod
    def create_driver(self):
        return self.driver_type(*self.driver_args,
                                signature_version='2',
                                host=self.default_host)

    def setUp(self):
        self.driver = self.create_driver()
        self.container = Container('test-container', {}, self.driver)
        self.object = Object('test-object', 1, 'hash', {},
                             'meta_data', self.container, self.driver)

    def test_connection_class_type(self):
        res = self.driver.connectionCls is DOSpacesConnectionAWS2
        self.assertTrue(res, 'driver.connectionCls does not match!')

    def test_connection_class_host(self):
        host = self.driver.connectionCls.host
        self.assertEqual(host, self.default_host)

    def test_container_enable_cdn_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.container.enable_cdn()

    def test_container_get_cdn_url_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.container.get_cdn_url()

    def test_object_enable_cdn_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.object.enable_cdn()

    def test_object_get_cdn_url_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.object.get_cdn_url()

    def test_invalid_signature_version(self):
        with self.assertRaises(ValueError):
            self.driver_type(*self.driver_args,
                             signature_version='3',
                             host=self.default_host)

    def test_invalid_region(self):
        with self.assertRaises(LibcloudError):
            self.driver_type(*self.driver_args,
                             region='atlantis',
                             host=self.default_host)


class DigitalOceanSpacesTests_v4(DigitalOceanSpacesTests):
    driver_type = DigitalOceanSpacesStorageDriver
    driver_args = STORAGE_S3_PARAMS
    default_host = 'nyc3.digitaloceanspaces.com'

    @classmethod
    def create_driver(self):
        return self.driver_type(*self.driver_args,
                                signature_version='4')

    def test_connection_class_type(self):
        res = self.driver.connectionCls is DOSpacesConnectionAWS4
        self.assertTrue(res, 'driver.connectionCls does not match!')

    def test_connection_class_host(self):
        host = self.driver.connectionCls.host
        self.assertEqual(host, self.default_host)

    def test_valid_regions(self):
        for region, hostname in DO_SPACES_HOSTS_BY_REGION.items():
            driver = self.driver_type(*self.driver_args,
                                      region=region)
            self.assertEqual(driver.connectionCls.host, hostname)
            self.assertTrue(driver.connectionCls.host.startswith(region))


class DigitalOceanSpacesDoubleInstanceTests(LibcloudTestCase):
    driver_type = DigitalOceanSpacesStorageDriver
    driver_args = STORAGE_S3_PARAMS
    default_host = 'nyc3.digitaloceanspaces.com'
    alt_host = 'ams3.digitaloceanspaces.com'

    def setUp(self):
        self.driver_v2 = self.driver_type(*self.driver_args,
                                          signature_version='2')
        self.driver_v4 = self.driver_type(*self.driver_args,
                                          signature_version='4',
                                          region='ams3')

    def test_connection_class_type(self):
        res = self.driver_v2.connectionCls is DOSpacesConnectionAWS2
        self.assertTrue(res, 'driver.connectionCls does not match!')

        res = self.driver_v4.connectionCls is DOSpacesConnectionAWS4
        self.assertTrue(res, 'driver.connectionCls does not match!')

        # Verify again that connection class hasn't been overriden when
        # instantiating a second driver class
        res = self.driver_v2.connectionCls is DOSpacesConnectionAWS2
        self.assertTrue(res, 'driver.connectionCls does not match!')

    def test_connection_class_host(self):
        host = self.driver_v2.connectionCls.host
        self.assertEqual(host, self.default_host)

        host = self.driver_v4.connectionCls.host
        self.assertEqual(host, self.alt_host)


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from libcloud.compute.types import Provider
from libcloud.compute.providers import get_driver
from libcloud.compute.base import NodeAuthPassword

ECSDriver = get_driver(Provider.ALIYUN_ECS)

region = 'cn-hangzhou'

your_access_key_id = ''
your_access_key_secret = ''
ecs = ECSDriver(your_access_key_id, your_access_key_secret, region=region)

sizes = ecs.list_sizes()
small = sizes[1]

locations = ecs.list_locations()
location = None
for each in locations:
    if each.id == region:
        location = each
        break
if location is None:
    print('could not find cn-qingdao location')
    sys.exit(-1)
print(location.name)

images = ecs.list_images()
print('Found %d images' % len(images))
for each in images:
    if 'ubuntu' in each.id.lower():
        image = each
        break
else:
    image = images[0]
print('Use image %s' % image)

sgs = ecs.ex_list_security_groups()
print('Found %d security groups' % len(sgs))
if len(sgs) == 0:
    sg = ecs.ex_create_security_group(description='test')
    print('Create security group %s' % sg)
else:
    sg = sgs[0].id
    print('Use security group %s' % sg)

nodes = ecs.list_nodes()
print('Found %d nodes' % len(nodes))
if len(nodes) == 0:
    print('Starting create a new node')
    data_disk = {
        'size': 5,
        'category': ecs.disk_categories.CLOUD,
        'disk_name': 'data_disk1',
        'delete_with_instance': True}

    auth = NodeAuthPassword('P@$$w0rd')

    ex_internet_charge_type = ecs.internet_charge_types.BY_TRAFFIC
    node = ecs.create_node(image=image, size=small, name='test',
                           ex_security_group_id=sg,
                           ex_internet_charge_type=ex_internet_charge_type,
                           ex_internet_max_bandwidth_out=1,
                           ex_data_disk=data_disk,
                           auth=auth)
    print('Created node %s' % node)
    nodes = ecs.list_nodes()

for each in nodes:
    print('Found node %s' % each)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'SLB_API_VERSION',
    'SLBDriver'
]

try:
    import simplejson as json
except ImportError:
    import json

from libcloud.common.aliyun import AliyunXmlResponse, SignedAliyunConnection
from libcloud.common.types import LibcloudError
from libcloud.loadbalancer.types import State
from libcloud.loadbalancer.base import Algorithm, Driver, LoadBalancer, Member
from libcloud.utils.misc import ReprMixin
from libcloud.utils.py3 import u
from libcloud.utils.xml import findattr, findtext, findall


SLB_API_VERSION = '2014-05-15'
SLB_API_HOST = 'slb.aliyuncs.com'
DEFAULT_SIGNATURE_VERSION = '1.0'


STATE_MAPPINGS = {
    'inactive': State.UNKNOWN,
    'active': State.RUNNING,
    'locked': State.PENDING
}


RESOURCE_EXTRA_ATTRIBUTES_MAP = {
    'balancer': {
        'create_timestamp': {
            'xpath': 'CreateTimeStamp',
            'transform_func': int
        },
        'address_type': {
            'xpath': 'AddressType',
            'transform_func': u
        },
        'region_id': {
            'xpath': 'RegionId',
            'transform_func': u
        },
        'region_id_alias': {
            'xpath': 'RegionIdAlias',
            'transform_func': u
        },
        'create_time': {
            'xpath': 'CreateTime',
            'transform_func': u
        },
        'master_zone_id': {
            'xpath': 'MasterZoneId',
            'transform_func': u
        },
        'slave_zone_id': {
            'xpath': 'SlaveZoneId',
            'transform_func': u
        },
        'network_type': {
            'xpath': 'NetworkType',
            'transform_func': u
        }
    }
}


SLB_SCHEDULER_TO_ALGORITHM = {
    'wrr': Algorithm.WEIGHTED_ROUND_ROBIN,
    'wlc': Algorithm.WEIGHTED_LEAST_CONNECTIONS
}


ALGORITHM_TO_SLB_SCHEDULER = {
    Algorithm.WEIGHTED_ROUND_ROBIN: 'wrr',
    Algorithm.WEIGHTED_LEAST_CONNECTIONS: 'wlc'
}


class SLBConnection(SignedAliyunConnection):
    api_version = SLB_API_VERSION
    host = SLB_API_HOST
    responseCls = AliyunXmlResponse
    service_name = 'slb'


class SLBLoadBalancerAttribute(object):
    """
    This class used to get listeners and backend servers related to a balancer
    listeners is a ``list`` of ``dict``, each element contains
    'ListenerPort' and 'ListenerProtocol' keys.
    backend_servers is a ``list`` of ``dict``, each element contains
    'ServerId' and 'Weight' keys.
    """
    def __init__(self, balancer, listeners, backend_servers, extra=None):
        self.balancer = balancer
        self.listeners = listeners or []
        self.backend_servers = backend_servers or []
        self.extra = extra or {}

    def is_listening(self, port):
        for listener in self.listeners:
            if listener.get('ListenerPort') == port:
                return True
        return False

    def is_attached(self, member):
        for server in self.backend_servers:
            if server.get('Serverid') == member.id:
                return True
        return False

    def __repr__(self):
        return ('<SLBLoadBalancerAttribute id=%s, ports=%s, servers=%s ...>' %
                (self.balancer.id, self.listeners, self.backend_servers))


class SLBLoadBalancerListener(ReprMixin, object):
    """
    Base SLB load balancer listener class
    """
    _repr_attributes = ['port', 'backend_port', 'scheduler', 'bandwidth']
    action = None
    option_keys = []

    def __init__(self, port, backend_port, algorithm, bandwidth, extra=None):
        self.port = port
        self.backend_port = backend_port
        self.scheduler = ALGORITHM_TO_SLB_SCHEDULER.get(algorithm, 'wrr')
        self.bandwidth = bandwidth
        self.extra = extra or {}

    @classmethod
    def create(cls, port, backend_port, algorithm, bandwidth, extra=None):
        return cls(port, backend_port, algorithm, bandwidth, extra=extra)

    def get_create_params(self):
        params = self.get_required_params()
        options = self.get_optional_params()
        options.update(params)
        return options

    def get_required_params(self):
        params = {'Action': self.action,
                  'ListenerPort': self.port,
                  'BackendServerPort': self.backend_port,
                  'Scheduler': self.scheduler,
                  'Bandwidth': self.bandwidth}
        return params

    def get_optional_params(self):
        options = {}
        for option in self.option_keys:
            if self.extra and option in self.extra:
                options[option] = self.extra[option]
        return options


class SLBLoadBalancerHttpListener(SLBLoadBalancerListener):
    """
    This class represents a rule to route http request to the backends.
    """
    action = 'CreateLoadBalancerHTTPListener'
    option_keys = ['XForwardedFor', 'StickySessionType', 'CookieTimeout',
                   'Cookie', 'HealthCheckDomain', 'HealthCheckURI',
                   'HealthCheckConnectPort', 'HealthyThreshold',
                   'UnhealthyThreshold', 'HealthCheckTimeout',
                   'HealthCheckInterval', 'HealthCheckHttpCode']

    def __init__(self, port, backend_port, algorithm, bandwidth,
                 sticky_session, health_check, extra=None):
        super(SLBLoadBalancerHttpListener, self).__init__(
            port, backend_port, algorithm, bandwidth, extra=extra)
        self.sticky_session = sticky_session
        self.health_check = health_check

    def get_required_params(self):
        params = super(SLBLoadBalancerHttpListener,
                       self).get_required_params()
        params['StickySession'] = self.sticky_session
        params['HealthCheck'] = self.health_check
        return params

    @classmethod
    def create(cls, port, backend_port, algorithm, bandwidth, extra={}):
        if 'StickySession' not in extra:
            raise AttributeError('StickySession is required')
        if 'HealthCheck' not in extra:
            raise AttributeError('HealthCheck is required')
        sticky_session = extra['StickySession']
        health_check = extra['HealthCheck']
        return cls(port, backend_port, algorithm, bandwidth, sticky_session,
                   health_check, extra=extra)


class SLBLoadBalancerHttpsListener(SLBLoadBalancerListener):
    """
    This class represents a rule to route https request to the backends.
    """
    action = 'CreateLoadBalancerHTTPSListener'
    option_keys = ['XForwardedFor', 'StickySessionType', 'CookieTimeout',
                   'Cookie', 'HealthCheckDomain', 'HealthCheckURI',
                   'HealthCheckConnectPort', 'HealthyThreshold',
                   'UnhealthyThreshold', 'HealthCheckTimeout',
                   'HealthCheckInterval', 'HealthCheckHttpCode']

    def __init__(self, port, backend_port, algorithm, bandwidth,
                 sticky_session, health_check, certificate_id, extra=None):
        super(SLBLoadBalancerHttpsListener, self).__init__(
            port, backend_port, algorithm, bandwidth, extra=extra)
        self.sticky_session = sticky_session
        self.health_check = health_check
        self.certificate_id = certificate_id

    def get_required_params(self):
        params = super(SLBLoadBalancerHttpsListener,
                       self).get_required_params()
        params['StickySession'] = self.sticky_session
        params['HealthCheck'] = self.health_check
        params['ServerCertificateId'] = self.certificate_id
        return params

    @classmethod
    def create(cls, port, backend_port, algorithm, bandwidth, extra={}):
        if 'StickySession' not in extra:
            raise AttributeError('StickySession is required')
        if 'HealthCheck' not in extra:
            raise AttributeError('HealthCheck is required')
        if 'ServerCertificateId' not in extra:
            raise AttributeError('ServerCertificateId is required')
        sticky_session = extra['StickySession']
        health_check = extra['HealthCheck']
        certificate_id = extra['ServerCertificateId']
        return cls(port, backend_port, algorithm, bandwidth, sticky_session,
                   health_check, certificate_id, extra=extra)


class SLBLoadBalancerTcpListener(SLBLoadBalancerListener):
    """
    This class represents a rule to route tcp request to the backends.
    """
    action = 'CreateLoadBalancerTCPListener'
    option_keys = ['PersistenceTimeout', 'HealthCheckType',
                   'HealthCheckDomain', 'HealthCheckURI',
                   'HealthCheckConnectPort', 'HealthyThreshold',
                   'UnhealthyThreshold', 'HealthCheckConnectTimeout',
                   'HealthCheckInterval', 'HealthCheckHttpCode']


class SLBLoadBalancerUdpListener(SLBLoadBalancerTcpListener):
    """
    This class represents a rule to route udp request to the backends.
    """
    action = 'CreateLoadBalancerUDPListener'
    option_keys = ['PersistenceTimeout', 'HealthCheckConnectPort',
                   'HealthyThreshold', 'UnhealthyThreshold',
                   'HealthCheckConnectTimeout', 'HealthCheckInterval']


class SLBServerCertificate(ReprMixin, object):
    _repr_attributes = ['id', 'name', 'fingerprint']

    def __init__(self, id, name, fingerprint):
        self.id = id
        self.name = name
        self.fingerprint = fingerprint


PROTOCOL_TO_LISTENER_MAP = {
    'http': SLBLoadBalancerHttpListener,
    'https': SLBLoadBalancerHttpsListener,
    'tcp': SLBLoadBalancerTcpListener,
    'udp': SLBLoadBalancerUdpListener
}


class SLBDriver(Driver):
    """
    Aliyun SLB load balancer driver.
    """
    name = 'Aliyun Server Load Balancer'
    website = 'https://www.aliyun.com/product/slb'
    connectionCls = SLBConnection
    path = '/'
    namespace = None

    _VALUE_TO_ALGORITHM_MAP = SLB_SCHEDULER_TO_ALGORITHM

    _ALGORITHM_TO_VALUE_MAP = ALGORITHM_TO_SLB_SCHEDULER

    def __init__(self, access_id, secret, region):
        super(SLBDriver, self).__init__(access_id, secret)
        self.region = region

    def list_protocols(self):
        return list(PROTOCOL_TO_LISTENER_MAP.keys())

    def list_balancers(self, ex_balancer_ids=None, ex_filters=None):
        """
        List all loadbalancers

        @inherits :class:`Driver.list_balancers`

        :keyword ex_balancer_ids: a list of balancer ids to filter results
                                  Only balancers which's id in this list
                                  will be returned
        :type ex_balancer_ids: ``list`` of ``str``

        :keyword ex_filters: attributes to filter results. Only balancers
                             which have all the desired attributes
                             and values will be returned
        :type ex_filters: ``dict``
        """

        params = {'Action': 'DescribeLoadBalancers',
                  'RegionId': self.region}
        if ex_balancer_ids and isinstance(ex_balancer_ids, list):
            params['LoadBalancerId'] = ','.join(ex_balancer_ids)

        if ex_filters and isinstance(ex_filters, dict):
            ex_filters.update(params)
            params = ex_filters
        resp_body = self.connection.request(self.path, params=params).object
        return self._to_balancers(resp_body)

    def create_balancer(self, name, port, protocol, algorithm, members,
                        ex_bandwidth=None, ex_internet_charge_type=None,
                        ex_address_type=None, ex_vswitch_id=None,
                        ex_master_zone_id=None, ex_slave_zone_id=None,
                        ex_client_token=None,
                        **kwargs):
        """
        Create a new load balancer instance

        @inherits: :class:`Driver.create_balancer`

        :keyword ex_bandwidth: The max bandwidth limit for `paybybandwidth`
                               internet charge type, in Mbps unit
        :type ex_bandwidth: ``int`` in range [1, 1000]

        :keyword ex_internet_charge_type: The internet charge type
        :type ex_internet_charge_type: a ``str`` of `paybybandwidth`
                                       or `paybytraffic`

        :keyword ex_address_type: The listening IP address type
        :type ex_address_type: a ``str`` of `internet` or `intranet`

        :keyword ex_vswitch_id: The vswitch id in a VPC network
        :type ex_vswitch_id: ``str``

        :keyword ex_master_zone_id: The id of the master availability zone
        :type ex_master_zone_id: ``str``

        :keyword ex_slave_zone_id: The id of the slave availability zone
        :type ex_slave_zone_id: ``str``

        :keyword ex_client_token: The token generated by client to
                                  identify requests
        :type ex_client_token: ``str``
        """

        # 1.Create load balancer
        params = {'Action': 'CreateLoadBalancer',
                  'RegionId': self.region}
        if name:
            params['LoadBalancerName'] = name
        if not port:
            raise AttributeError('port is required')
        if not protocol:
            # NOTE(samsong8610): Use http listener as default
            protocol = 'http'
        if protocol not in PROTOCOL_TO_LISTENER_MAP:
            raise AttributeError('unsupport protocol %s' % protocol)

        # Bandwidth in range [1, 1000] Mbps
        bandwidth = -1
        if ex_bandwidth:
            try:
                bandwidth = int(ex_bandwidth)
            except ValueError:
                raise AttributeError('ex_bandwidth should be a integer in '
                                     'range [1, 1000].')
            params['Bandwidth'] = bandwidth

        if ex_internet_charge_type:
            if ex_internet_charge_type.lower() == 'paybybandwidth':
                if bandwidth == -1:
                    raise AttributeError('PayByBandwidth internet charge type'
                                         ' need ex_bandwidth be set')
            params['InternetChargeType'] = ex_internet_charge_type

        if ex_address_type:
            if ex_address_type.lower() not in ('internet', 'intranet'):
                raise AttributeError('ex_address_type should be "internet" '
                                     'or "intranet"')
            params['AddressType'] = ex_address_type

        if ex_vswitch_id:
            params['VSwitchId'] = ex_vswitch_id

        if ex_master_zone_id:
            params['MasterZoneId'] = ex_master_zone_id
        if ex_slave_zone_id:
            params['SlaveZoneId'] = ex_slave_zone_id

        if ex_client_token:
            params['ClientToken'] = ex_client_token

        if members and isinstance(members, list):
            backend_ports = [member.port for member in members]
            if len(set(backend_ports)) != 1:
                raise AttributeError('the ports of members should be unique')
            # NOTE(samsong8610): If members do not provide backend port,
            #                    default to listening port
            backend_port = backend_ports[0] or port
        else:
            backend_port = port

        balancer = None
        try:
            resp_body = self.connection.request(self.path, params).object
            balancer = self._to_balancer(resp_body)
            balancer.port = port

            # 2.Add backend servers
            if members is None:
                members = []
            for member in members:
                self.balancer_attach_member(balancer, member)
            # 3.Create listener
            # NOTE(samsong8610): Assume only create a listener which uses all
            #                    the bandwidth.
            self.ex_create_listener(balancer, backend_port, protocol,
                                    algorithm, bandwidth, **kwargs)
            self.ex_start_listener(balancer, port)
            return balancer
        except Exception as e:
            if balancer is not None:
                try:
                    self.destroy_balancer(balancer)
                except Exception:
                    pass
            raise e

    def destroy_balancer(self, balancer):
        params = {'Action': 'DeleteLoadBalancer',
                  'LoadBalancerId': balancer.id}
        resp = self.connection.request(self.path, params)
        return resp.success()

    def get_balancer(self, balancer_id):
        balancers = self.list_balancers(ex_balancer_ids=[balancer_id])
        if len(balancers) != 1:
            raise LibcloudError('could not find load balancer with id %s' %
                                balancer_id)
        return balancers[0]

    def balancer_attach_compute_node(self, balancer, node):
        if len(node.public_ips) > 0:
            ip = node.public_ips[0]
        else:
            ip = node.private_ips[0]
        member = Member(id=node.id, ip=ip, port=balancer.port)
        return self.balancer_attach_member(balancer, member)

    def balancer_attach_member(self, balancer, member):
        params = {'Action': 'AddBackendServers',
                  'LoadBalancerId': balancer.id}
        if member and isinstance(member, Member):
            params['BackendServers'] = self._to_servers_json([member])
        self.connection.request(self.path, params)
        return member

    def balancer_detach_member(self, balancer, member):
        params = {'Action': 'RemoveBackendServers',
                  'LoadBalancerId': balancer.id}
        if member and isinstance(member, Member):
            params['BackendServers'] = self._list_to_json([member.id])
        self.connection.request(self.path, params)
        return member

    def balancer_list_members(self, balancer):
        attribute = self.ex_get_balancer_attribute(balancer)
        members = [Member(server['ServerId'], None, None, balancer=balancer,
                          extra={'Weight': server['Weight']})
                   for server in attribute.backend_servers]
        return members

    def ex_get_balancer_attribute(self, balancer):
        """
        Get balancer attribute

        :param balancer: the balancer to get attribute
        :type balancer: ``LoadBalancer``

        :return: the balancer attribute
        :rtype: ``SLBLoadBalancerAttribute``
        """

        params = {'Action': 'DescribeLoadBalancerAttribute',
                  'LoadBalancerId': balancer.id}
        resp_body = self.connection.request(self.path, params).object
        attribute = self._to_balancer_attribute(resp_body)
        return attribute

    def ex_list_listeners(self, balancer):
        """
        Get all listener related to the given balancer

        :param balancer: the balancer to list listeners
        :type balancer: ``LoadBalancer``

        :return: a list of listeners
        :rtype: ``list`` of ``SLBLoadBalancerListener``
        """

        attribute = self.ex_get_balancer_attribute(balancer)
        listeners = [SLBLoadBalancerListener(each['ListenerPort'], None,
                                             None, None)
                     for each in attribute.listeners]
        return listeners

    def ex_create_listener(self, balancer, backend_port, protocol, algorithm,
                           bandwidth, **kwargs):
        """
        Create load balancer listening rule.

        :param balancer: the balancer which the rule belongs to.
                         The listener created will listen on the port of the
                         the balancer as default. 'ListenerPort' in kwargs
                         will *OVERRIDE* it.
        :type balancer: ``LoadBalancer``

        :param backend_port: the backend server port
        :type backend_port: ``int``

        :param protocol: the balancer protocol, default to http
        :type protocol: ``str``

        :param algorithm: the balancer routing algorithm
        :type algorithm: ``Algorithm``

        :param bandwidth: the listener bandwidth limits
        :type bandwidth: ``str``

        :return: the created listener
        :rtype: ``SLBLoadBalancerListener``
        """

        cls = PROTOCOL_TO_LISTENER_MAP.get(protocol,
                                           SLBLoadBalancerHttpListener)
        if 'ListenerPort' in kwargs:
            port = kwargs['ListenerPort']
        else:
            port = balancer.port
        listener = cls.create(port, backend_port, algorithm,
                              bandwidth, extra=kwargs)
        params = listener.get_create_params()
        params['LoadBalancerId'] = balancer.id
        params['RegionId'] = self.region
        resp = self.connection.request(self.path, params)
        return resp.success()

    def ex_start_listener(self, balancer, port):
        """
        Start balancer's listener listening the given port.

        :param balancer: a load balancer
        :type balancer: ``LoadBalancer``

        :param port: listening port
        :type port: ``int``

        :return: whether operation is success
        :rtype: ``bool``
        """

        params = {'Action': 'StartLoadBalancerListener',
                  'LoadBalancerId': balancer.id,
                  'ListenerPort': port}
        resp = self.connection.request(self.path, params)
        return resp.success()

    def ex_stop_listener(self, balancer, port):
        """
        Stop balancer's listener listening the given port.

        :param balancer: a load balancer
        :type balancer: ``LoadBalancer``

        :param port: listening port
        :type port: ``int``

        :return: whether operation is success
        :rtype: ``bool``
        """

        params = {'Action': 'StopLoadBalancerListener',
                  'LoadBalancerId': balancer.id,
                  'ListenerPort': port}
        resp = self.connection.request(self.path, params)
        return resp.success()

    def ex_upload_certificate(self, name, server_certificate,
                              private_key):
        """
        Upload certificate and private key for https load balancer listener

        :param name: the certificate name
        :type name: ``str``

        :param server_certificate: the content of the certificate to upload
                                   in PEM format
        :type server_certificate: ``str``

        :param private_key: the content of the private key to upload
                            in PEM format
        :type private_key: ``str``

        :return: new created certificate info
        :rtype: ``SLBServerCertificate``
        """

        params = {'Action': 'UploadServerCertificate',
                  'RegionId': self.region,
                  'ServerCertificate': server_certificate,
                  'PrivateKey': private_key}
        if name:
            params['ServerCertificateName'] = name
        resp_body = self.connection.request(self.path, params).object
        return self._to_server_certificate(resp_body)

    def ex_list_certificates(self, certificate_ids=[]):
        """
        List all server certificates

        :param certificate_ids: certificate ids to filter results
        :type certificate_ids: ``str``

        :return: certificates
        :rtype: ``SLBServerCertificate``
        """

        params = {'Action': 'DescribeServerCertificates',
                  'RegionId': self.region}
        if certificate_ids and isinstance(certificate_ids, list):
            params['ServerCertificateId'] = ','.join(certificate_ids)

        resp_body = self.connection.request(self.path, params).object
        cert_elements = findall(resp_body,
                                'ServerCertificates/ServerCertificate',
                                namespace=self.namespace)
        certificates = [self._to_server_certificate(el)
                        for el in cert_elements]
        return certificates

    def ex_delete_certificate(self, certificate_id):
        """
        Delete the given server certificate

        :param certificate_id: the id of the certificate to delete
        :type certificate_id: ``str``

        :return: whether process is success
        :rtype: ``bool``
        """

        params = {'Action': 'DeleteServerCertificate',
                  'RegionId': self.region,
                  'ServerCertificateId': certificate_id}
        resp = self.connection.request(self.path, params)
        return resp.success()

    def ex_set_certificate_name(self, certificate_id, name):
        """
        Set server certificate name.

        :param certificate_id: the id of the server certificate to update
        :type certificate_id: ``str``

        :param name: the new name
        :type name: ``str``

        :return: whether updating is success
        :rtype: ``bool``
        """

        params = {'Action': 'SetServerCertificateName',
                  'RegionId': self.region,
                  'ServerCertificateId': certificate_id,
                  'ServerCertificateName': name}
        resp = self.connection.request(self.path, params)
        return resp.success()

    def _to_balancers(self, element):
        xpath = 'LoadBalancers/LoadBalancer'
        return [self._to_balancer(el)
                for el in findall(element=element, xpath=xpath,
                                  namespace=self.namespace)]

    def _to_balancer(self, el):
        _id = findtext(element=el, xpath='LoadBalancerId',
                       namespace=self.namespace)
        name = findtext(element=el, xpath='LoadBalancerName',
                        namespace=self.namespace)
        status = findtext(element=el, xpath='LoadBalancerStatus',
                          namespace=self.namespace)
        state = STATE_MAPPINGS.get(status, State.UNKNOWN)
        address = findtext(element=el, xpath='Address',
                           namespace=self.namespace)
        extra = self._get_extra_dict(
            el, RESOURCE_EXTRA_ATTRIBUTES_MAP['balancer'])

        balancer = LoadBalancer(id=_id, name=name, state=state, ip=address,
                                port=None, driver=self, extra=extra)
        return balancer

    def _create_list_params(self, params, items, label):
        """
        return parameter list
        """
        if isinstance(items, str):
            items = [items]
        for index, item in enumerate(items):
            params[label % (index + 1)] = item
        return params

    def _get_extra_dict(self, element, mapping):
        """
        Extract attributes from the element based on rules provided in the
        mapping dictionary.

        :param      element: Element to parse the values from.
        :type       element: xml.etree.ElementTree.Element.

        :param      mapping: Dictionary with the extra layout
        :type       node: :class:`Node`

        :rtype: ``dict``
        """
        extra = {}
        for attribute, values in mapping.items():
            transform_func = values['transform_func']
            value = findattr(element=element,
                             xpath=values['xpath'],
                             namespace=self.namespace)
            if value:
                try:
                    extra[attribute] = transform_func(value)
                except Exception:
                    extra[attribute] = None
            else:
                extra[attribute] = value

        return extra

    def _to_servers_json(self, members):
        servers = []
        for each in members:
            server = {'ServerId': each.id,
                      'Weight': '100'}
            if 'Weight' in each.extra:
                server['Weight'] = each.extra['Weight']
            servers.append(server)
        try:
            return json.dumps(servers)
        except Exception:
            raise AttributeError('could not convert member to backend server')

    def _to_balancer_attribute(self, element):
        balancer = self._to_balancer(element)
        port_proto_elements = findall(
            element, 'ListenerPortsAndProtocol/ListenerPortAndProtocol',
            namespace=self.namespace)
        if len(port_proto_elements) > 0:
            listeners = [self._to_port_and_protocol(el)
                         for el in port_proto_elements]
        else:
            port_elements = findall(element, 'ListenerPorts/ListenerPort',
                                    namespace=self.namespace)
            listeners = [{'ListenerPort': el.text, 'ListenerProtocol': 'http'}
                         for el in port_elements]
        server_elements = findall(element,
                                  'BackendServers/BackendServer',
                                  namespace=self.namespace)
        backend_servers = [self._to_server_and_weight(el)
                           for el in server_elements]
        return SLBLoadBalancerAttribute(balancer, listeners, backend_servers)

    def _to_port_and_protocol(self, el):
        port = findtext(el, 'ListenerPort', namespace=self.namespace)
        protocol = findtext(el, 'ListenerProtocol', namespace=self.namespace)
        return {'ListenerPort': port, 'ListenerProtocol': protocol}

    def _to_server_and_weight(self, el):
        server_id = findtext(el, 'ServerId', namespace=self.namespace)
        weight = findtext(el, 'Weight', namespace=self.namespace)
        return {'ServerId': server_id, 'Weight': weight}

    def _to_server_certificate(self, el):
        _id = findtext(el, 'ServerCertificateId', namespace=self.namespace)
        name = findtext(el, 'ServerCertificateName', namespace=self.namespace)
        fingerprint = findtext(el, 'Fingerprint', namespace=self.namespace)
        return SLBServerCertificate(id=_id, name=name,
                                    fingerprint=fingerprint)

    def _list_to_json(self, value):
        try:
            return json.dumps(value)
        except Exception:
            return '[]'
